# Copyright The Orthority Contributors.
#
# This file is part of Orthority.
#
# Orthority is free software: you can redistribute it and/or modify it under the terms of the GNU
# Affero General Public License as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# Orthority is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Orthority.
# If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import logging
import os
from contextlib import ExitStack
from functools import partial
from itertools import product
from pathlib import Path

import numpy as np
import pytest
import rasterio as rio
from rasterio.enums import Resampling, MaskFlags, ColorInterp
from rasterio.vrt import WarpedVRT
from rasterio.windows import Window

from orthority import common
from orthority.enums import Interp, Compress, Driver
from orthority.errors import OrthorityError, OrthorityWarning
from orthority.pan_sharp import PanSharpen
from tests.conftest import create_profile

logger = logging.getLogger(__name__)


@pytest.fixture(scope='session')
def pan_sharpen(pan_file: Path, ms_file: Path) -> PanSharpen:
    """A pan-sharpener."""
    return PanSharpen(pan_file, ms_file)


@pytest.fixture(scope='session')
def mask_pan_file(
    pan_file: Path, webmerc_crs: str, tmp_path_factory: pytest.TempPathFactory
) -> Path:
    """Panchromatic image with georeferencing and an internal mask defining a buffer of invalid
    pixels.
    """
    # read pan_file
    with rio.open(pan_file, 'r') as src_im:
        src_array = src_im.read()

    # pad src_array and create georeferenced profile
    pwidth = 100
    pval = 0
    out_array = np.pad(
        src_array, ((0, 0), (pwidth, pwidth), (pwidth, pwidth)), constant_values=pval
    )
    transform = rio.Affine(2, 0, 10, 0, -2, 20)
    out_profile = create_profile(out_array, crs=webmerc_crs, transform=transform)

    # write to output file
    out_file = tmp_path_factory.mktemp('data').joinpath('mask_pan.tif')
    with rio.open(out_file, 'w', **out_profile) as out_im:
        out_im.write(out_array)
        out_im.write_mask((out_array != pval).all(axis=0))
    return out_file


@pytest.fixture(scope='session')
def mask_ms_file(ms_file: Path, webmerc_crs: str, tmp_path_factory: pytest.TempPathFactory) -> Path:
    """Multispectral image with georeferencing and an internal mask defining a buffer of invalid
    pixels.
    """
    # read ms_file
    with rio.open(ms_file, 'r') as src_im:
        src_array = src_im.read()

    # pad src_array  and create georeferenced profile that shares bounds with mask_pan_file
    scale = 4  # scale relative to mask_pan_file
    pwidth = int(100 / scale)
    pval = 0
    out_array = np.pad(
        src_array, ((0, 0), (pwidth, pwidth), (pwidth, pwidth)), constant_values=pval
    )
    transform = rio.Affine(2, 0, 10, 0, -2, 20) * rio.Affine.scale(scale)
    out_profile = create_profile(out_array, crs=webmerc_crs, transform=transform)

    # write to output file
    out_file = tmp_path_factory.mktemp('data').joinpath('mask_ms.tif')
    with rio.open(out_file, 'w', **out_profile) as out_im:
        out_im.write(out_array)
        out_im.write_mask((out_array != pval).all(axis=0))
    return out_file


@pytest.fixture(scope='session')
def ms_neg_band_file(ms_file: Path, tmp_path_factory: pytest.TempPathFactory) -> Path:
    """Multispectral image that is ``ms_file`` with a fourth band added that is the inverse of the
    first band.  Also adds a non-standard ``colorinterp``.
    """
    # read ms_file
    with rio.open(ms_file, 'r') as src_im:
        src_array = src_im.read(out_dtype='int16')

    # create 4 band array with 4th band as inverse of first band
    out_array = np.concatenate((src_array, [src_array[0] * -1]), axis=0)

    # write to output file
    out_profile = create_profile(out_array)
    out_file = tmp_path_factory.mktemp('data').joinpath('ms_neg_band.tif')
    with rio.open(out_file, 'w', **out_profile) as out_im:
        out_im.colorinterp = [
            ColorInterp.blue,
            ColorInterp.green,
            ColorInterp.red,
            ColorInterp.undefined,
        ]
        out_im.write(out_array)
    return out_file


def test_init_profiles(pan_file: Path, ms_file: Path):
    """Basic test of pan / MS reprojection profiles."""
    pan_sharp = PanSharpen(pan_file, ms_file)

    req_keys = ['crs', 'src_transform', 'transform', 'width', 'height']
    for profile_key in ['pan_to_pan', 'pan_to_ms', 'ms_to_ms', 'ms_to_pan']:
        assert profile_key in pan_sharp._profiles
        profile = pan_sharp._profiles[profile_key]
        assert all([k in profile for k in req_keys])
        assert profile['dtype'] == PanSharpen._working_dtype
        assert common.nan_equals(profile['nodata'], PanSharpen._working_nodata)
        assert profile['num_threads'] == os.cpu_count()


def test_init_res_error(pan_file: Path, ms_file: Path):
    """Test initialisation raises an error when the pan resolution is larger than the multispectral
    resolution.
    """
    with pytest.raises(OrthorityError) as ex:
        _ = PanSharpen(ms_file, pan_file)
    assert 'resolution' in str(ex.value)


def test_init_bounds_error(odm_dem_file: Path, ngi_dem_file: Path):
    """Test initialisation raises an error when pan and multispectral bounds don't overlap."""
    with pytest.raises(OrthorityError) as ex:
        _ = PanSharpen(odm_dem_file, ngi_dem_file)
    assert 'bounds' in str(ex.value)


def _test_pan_sharp_file(
    out_file: Path, ms_file: Path | rio.DatasetReader, ms_indexes: tuple = None
):
    """Test the pan-sharpened ``out_file`` by reprojecting it to the grid of the given
    multispectral ``ms_file`` and comparing.
    """
    with common.OpenRaster(ms_file, 'r') as ms_im, common.OpenRaster(out_file, 'r') as out_im_:
        # prepare parameters for reprojecting output to MS grid
        if (ms_im.crs is None and ms_im.transform.is_identity) or (
            out_im_.crs is None and out_im_.transform.is_identity
        ):
            # MS or output not georeferenced - assume bounds match
            ms_crs = out_im_.crs
            scale = np.array(out_im_.shape) / ms_im.shape
            ms_transform = out_im_.transform * rio.Affine.scale(*scale[::-1])
            ms_win = Window(0, 0, ms_im.width, ms_im.height)
        else:
            # MS and output georeferenced
            ms_crs = ms_im.crs
            ms_transform = ms_im.transform
            # find MS window corresponding to output bounds
            ms_win = common.expand_window_to_grid(ms_im.window(*out_im_.bounds))

        profile = dict(
            crs=ms_crs,
            transform=ms_transform,
            width=ms_im.width,
            height=ms_im.height,
            resampling=Resampling.average,
        )

        # read reprojected output
        with WarpedVRT(out_im_, **profile) as out_im:
            out_array = out_im.read(window=ms_win, masked=True, out_dtype='float32')

        # read MS window corresponding to output bounds
        ms_indexes = ms_indexes or [
            bi + 1 for bi in range(ms_im.count) if ms_im.colorinterp[bi] != ColorInterp.alpha
        ]
        ms_array = ms_im.read(ms_indexes, window=ms_win, masked=True, out_dtype='float32')

    # compare output to MS
    assert np.all(out_array.mask == ms_array.mask)
    abs_err = np.ma.abs(out_array - ms_array)
    assert np.ma.all(abs_err.mean(axis=(1, 2)) < 1)
    assert np.ma.all(abs_err.std(axis=(1, 2)) < 1)


@pytest.mark.parametrize(
    'pan_profile, ms_profile',
    [
        # pan & MS not georeferenced
        (dict(), dict()),
        # pan georeferenced
        (
            dict(
                crs='EPSG:3857', src_transform=(t := rio.Affine(2, 0, 10, 0, -2, 20)), transform=t
            ),
            dict(),
        ),
        # MS georeferenced
        (
            dict(),
            dict(crs='EPSG:3857', src_transform=t, transform=t),
        ),
        # pan & MS georeferenced with matching bounds
        (
            dict(crs='EPSG:3857', src_transform=t, transform=t),
            dict(crs='EPSG:3857', src_transform=(ms_t := t * rio.Affine.scale(4)), transform=ms_t),
        ),
        # pan & MS georeferenced with different bounds
        (
            dict(crs='EPSG:3857', src_transform=t, transform=t, width=1200, height=800),
            dict(
                crs='EPSG:3857',
                src_transform=ms_t,
                transform=ms_t * rio.Affine.translation(50, 50),
                width=342 - 50,
                height=228 - 50,
            ),
        ),
    ],
)
def test_georef(pan_file: Path, ms_file: Path, pan_profile: dict, ms_profile: dict, tmp_path: Path):
    """Test pan-sharpening using WarpedVRT to simulate different combinations of pan & MS
    georeferencing and bounds (all without masks).
    """
    out_file = tmp_path.joinpath('pan_sharp.tif')
    with rio.open(pan_file, 'r') as pan_im_, rio.open(ms_file, 'r') as ms_im_:
        # set profile widths & heights if they are not set already
        pan_profile['width'] = pan_profile.get('width', pan_im_.width)
        pan_profile['height'] = pan_profile.get('height', pan_im_.height)
        ms_profile['width'] = ms_profile.get('width', ms_im_.width)
        ms_profile['height'] = ms_profile.get('height', ms_im_.height)

        # simulate pan & MS georeferencing, and pan-sharpen
        with WarpedVRT(pan_im_, **pan_profile) as pan_im, WarpedVRT(ms_im_, **ms_profile) as ms_im:
            pan_sharp = PanSharpen(pan_im, ms_im)
            pan_sharp.process(out_file, weights=(1, 1, 1), compress='deflate')
            assert out_file.exists()

            # reproject output to MS grid and compare
            _test_pan_sharp_file(out_file, ms_im)


@pytest.mark.parametrize('vrt_profile', [None, dict(nodata=0), dict(add_alpha=True)])
def test_mask(mask_pan_file: Path, mask_ms_file: Path, tmp_path: Path, vrt_profile: dict):
    """Test masking in the pan-sharpened image with pan & MS images that have internal,
    nodata and alpha masks, that mask invalid pixels inside the common bounds.
    """
    with ExitStack() as exit_stack:
        pan_im = exit_stack.enter_context(rio.open(mask_pan_file, 'r'))
        ms_im = exit_stack.enter_context(rio.open(mask_ms_file, 'r'))

        if vrt_profile:
            pan_im = exit_stack.enter_context(WarpedVRT(pan_im, **vrt_profile))
            ms_im = exit_stack.enter_context(WarpedVRT(ms_im, **vrt_profile))

        pan_sharp = PanSharpen(pan_im, ms_im)
        out_file = tmp_path.joinpath('pan_sharp.tif')
        pan_sharp.process(out_file, weights=(1, 1, 1), compress='deflate')
        assert out_file.exists()

        # reproject output to MS grid and compare
        _test_pan_sharp_file(out_file, ms_im)


def test_stats(pan_file: Path, ms_file: Path, monkeypatch: pytest.MonkeyPatch):
    """Test block-by-block ``PanSharpen._get_stats()`` against full band reference values."""
    pan_sharp = PanSharpen(pan_file, ms_file)

    with rio.open(ms_file, 'r') as ms_im, rio.open(pan_file, 'r') as pan_im_:
        # find reference full band stats
        ms_array = ms_im.read(out_dtype='float64')
        with WarpedVRT(
            pan_im_, **pan_sharp._profiles['pan_to_ms'], resampling=Resampling.average
        ) as pan_im:
            pan_array = pan_im.read()
        pan_ms_array = np.concatenate(
            (pan_array.reshape(1, -1), ms_array.reshape(ms_array.shape[0], -1)), axis=0
        )
        ref_means = pan_ms_array.mean(axis=1)
        ref_cov = np.cov(pan_ms_array)

        # find test block-by-block stats
        pan_sharp = PanSharpen(pan_file, ms_file)
        # patch common.block_windows to use smaller block size (forces >> 1 block)
        monkeypatch.setattr(
            common, 'block_windows', partial(common.block_windows, block_shape=(64, 64))
        )
        test_means, test_cov = pan_sharp._get_stats(pan_im_, ms_im, 1, ms_im.indexes, {})

    assert test_means == pytest.approx(ref_means, abs=1e-6)
    assert test_cov == pytest.approx(ref_cov, abs=1e-6)


def test_weights_auto(pan_file: Path, ms_file: Path):
    """Test estimated MS to pan weights returned by ``PanSharpen._get_params()`` against expected
    values.
    """
    pan_sharp = PanSharpen(pan_file, ms_file)
    with rio.open(ms_file, 'r') as ms_im, rio.open(pan_file, 'r') as pan_im:
        means, cov = pan_sharp._get_stats(pan_im, ms_im, 1, ms_im.indexes, {})
    params = pan_sharp._get_params(means, cov, None)

    # assumes pan == sum(MS bands) exactly
    assert params['weights'] == pytest.approx(1 / 3, abs=0.01)
    assert np.sum(params['weights']) == pytest.approx(1, abs=1e-6)


@pytest.mark.parametrize('weights', [(1, 1, 1), (1, 2, 3), (-1, 0, 1)])
def test_weights_user(pan_file: Path, ms_file: Path, weights: tuple | None):
    """Test user provided MS to pan weights are clipped, normalised and passed through
    ``PanSharpen._get_params()``.
    """
    pan_sharp = PanSharpen(pan_file, ms_file)
    with rio.open(ms_file, 'r') as ms_im, rio.open(pan_file, 'r') as pan_im:
        means, cov = pan_sharp._get_stats(pan_im, ms_im, 1, ms_im.indexes, {})
    params = pan_sharp._get_params(means, cov, weights)

    weights = np.array(weights).clip(0, None)
    weights = weights / weights.sum()
    assert np.all(params['weights'] == weights)


def test_weights_negative(pan_file: Path, ms_neg_band_file: Path):
    """Test behaviour of ``PanSharpen._get_params()`` MS to pan weights estimation when estimated
    weights contain negative values.
    """
    pan_sharp = PanSharpen(pan_file, ms_neg_band_file)
    with rio.open(ms_neg_band_file, 'r') as ms_im, rio.open(pan_file, 'r') as pan_im:
        means, cov = pan_sharp._get_stats(pan_im, ms_im, 1, ms_im.indexes, {})

    # get weights, testing a warning is raised
    with pytest.warns(OrthorityWarning) as record:
        params = pan_sharp._get_params(means, cov, None)
    assert len(record) == 1
    assert 'Weights contain negative value(s)' in record[0].message.args[0]

    # test weight of negative band has been clipped to 0
    assert params['weights'][-1] == 0

    # test remaining weights have been re-estimated and normalised
    assert params['weights'][:-1] == pytest.approx(1 / 3, abs=0.01)
    assert np.sum(params['weights']) == pytest.approx(1, abs=1e-6)


@pytest.mark.parametrize('weights', [(1, 1, 1), (1, 2, 3)])
def test_gs_coeffs(pan_file: Path, ms_file: Path, weights: tuple):
    """Test ``PanSharpen._get_params()`` Gram-Schmidt coefficients against full band reference
    values.
    """
    working_dtype = PanSharpen._working_dtype
    weights = np.array(weights) / np.sum(weights)

    # find reference coefficients using full band approach
    def gs_foward(ms_array: np.array, weights: np.array) -> list[np.array]:
        """Full band forward GS transform from https://patents.google.com/patent/US6011875A/en."""
        gs_pan = weights.dot(ms_array)
        gs_array = np.zeros((ms_array.shape[0] + 1, ms_array.shape[1]), dtype=working_dtype)
        gs_array[0] = gs_pan
        gs_coeffs = []
        means = ms_array.mean(axis=1)

        for bi in range(0, ms_array.shape[0]):
            c = np.cov(ms_array[bi], gs_array[: bi + 1])
            phi = c[0, 1:] / np.diag(c)[1:]
            gs_array[bi + 1] = ms_array[bi] - means[bi] - phi.dot(gs_array[: bi + 1])
            gs_coeffs.append(phi)

        return gs_coeffs

    with rio.open(ms_file, 'r') as ms_im:
        ms_array = ms_im.read(out_dtype=working_dtype)
    ms_array = ms_array.reshape(ms_array.shape[0], -1)
    ref_coeffs = gs_foward(ms_array, weights)

    # find test coefficients using PanSharpen
    pan_sharp = PanSharpen(pan_file, ms_file)
    with rio.open(ms_file, 'r') as ms_im, rio.open(pan_file, 'r') as pan_im:
        means, cov = pan_sharp._get_stats(pan_im, ms_im, 1, ms_im.indexes, {})
    params = pan_sharp._get_params(means, cov, weights)

    for test_coeff, ref_coeff in zip(params['coeffs'], ref_coeffs):
        assert test_coeff == pytest.approx(ref_coeff, abs=1e-6)


@pytest.mark.parametrize('weights', [(1, 1, 1), (1, 2, 3)])
def test_pan_norm(pan_file: Path, ms_file: Path, weights: tuple):
    """Test ``PanSharpen._get_params()`` pan normalisation parameters against full band reference
    values.
    """
    working_dtype = PanSharpen._working_dtype
    weights = np.array(weights) / np.sum(weights)
    pan_sharp = PanSharpen(pan_file, ms_file)

    # find reference parameters using full band approach
    def get_pan_norm(pan_array: np.ndarray, ms_array: np.ndarray, weights: np.ndarray):
        """Return pan normalisation parameters for given arrays."""
        pan_mean, pan_std = pan_array.mean(), pan_array.std()
        pan_sim_array = np.dot(weights, ms_array)
        pan_sim_mean, pan_sim_std = pan_sim_array.mean(), pan_sim_array.std()
        gain = pan_sim_std / pan_std
        bias = pan_sim_mean - (gain * pan_mean)
        return gain, bias

    with rio.open(ms_file, 'r') as ms_im, rio.open(pan_file, 'r') as pan_im_:
        ms_array = ms_im.read(out_dtype=working_dtype)
        ms_array = ms_array.reshape(ms_array.shape[0], -1)

        with WarpedVRT(
            pan_im_, **pan_sharp._profiles['pan_to_ms'], resampling=Resampling.average
        ) as pan_im:
            pan_array = pan_im.read()
            pan_array = pan_array.reshape(1, -1)

        ref_gain, ref_bias = get_pan_norm(pan_array, ms_array, weights)

        # find test parameters using PanSharpen
        means, cov = pan_sharp._get_stats(pan_im_, ms_im, 1, ms_im.indexes, {})
        params = pan_sharp._get_params(means, cov, weights)

    assert params['gain'] == pytest.approx(ref_gain, abs=1e-6)
    assert params['bias'] == pytest.approx(ref_bias, abs=1e-6)


def test_process_pan_index_error(pan_sharpen: PanSharpen, tmp_path: Path):
    """Test ``PanSharpen.process()`` raises an error when ``pan_index`` is invalid."""
    out_file = tmp_path.joinpath('pan_sharp.tif')
    with pytest.raises(OrthorityError) as ex:
        pan_sharpen.process(out_file, pan_index=2)
    assert 'Pan index' in str(ex)


@pytest.mark.parametrize('ms_indexes, weights', [((1, 2, 3), (1, 1, 1)), ((3, 2, 1), (1, 1, 1))])
def test_process_ms_indexes(
    pan_file: Path, ms_file: Path, tmp_path: Path, ms_indexes: tuple, weights: tuple
):
    """Test pan-sharpened bands are correctly defined by the ``PanSharpen.process()``
    ``ms_indexes`` argument.
    """
    # note that for this test to pass, ms_indexes and weights must give a weighted sum of indexed
    # MS bands, that equals the plain sum of MS bands (as pan_file==sum(ms_file bands))
    pan_sharp = PanSharpen(pan_file, ms_file)
    out_file = tmp_path.joinpath('pan_sharp.tif')
    pan_sharp.process(out_file, ms_indexes=ms_indexes, weights=weights, compress='deflate')
    assert out_file.exists()

    _test_pan_sharp_file(out_file, ms_file, ms_indexes=ms_indexes)


def test_process_ms_indexes_error(pan_sharpen: PanSharpen, tmp_path: Path):
    """Test ``PanSharpen.process()`` raises an error when ``ms_indexes`` is invalid."""
    out_file = tmp_path.joinpath('pan_sharp.tif')
    for ms_indexes in [[0], [4]]:
        with pytest.raises(OrthorityError) as ex:
            pan_sharpen.process(out_file, ms_indexes=ms_indexes)
        assert 'Multispectral indexes' in str(ex)


def test_process_weights(pan_file: Path, ms_file: Path, tmp_path: Path):
    """Test the ``PanSharpen.process()`` ``weights`` argument by comparing pan-sharpened images
    with different weightings.
    """
    out_arrays = []
    for wi, weights in enumerate([(1, 2, 3), (3, 2, 1)]):
        pan_sharp = PanSharpen(pan_file, ms_file)
        out_file = tmp_path.joinpath(f'pan_sharp_{wi}.tif')
        pan_sharp.process(out_file, weights=weights, compress='deflate')
        assert out_file.exists()

        with rio.open(out_file, 'r') as out_im:
            out_arrays.append(out_im.read())

    assert np.any(out_arrays[0] != out_arrays[1])


def test_process_weights_error(pan_sharpen: PanSharpen, tmp_path: Path):
    """Test ``PanSharpen.process()`` raises an error when ``weights`` is invalid."""
    out_file = tmp_path.joinpath('pan_sharp.tif')
    with pytest.raises(OrthorityError) as ex:
        pan_sharpen.process(out_file, weights=(1, 1))
    assert 'weights' in str(ex) and 'same number' in str(ex)
    with pytest.raises(OrthorityError) as ex:
        pan_sharpen.process(out_file, weights=(1, 1, -1))
    assert 'Weight values' in str(ex)


@pytest.mark.parametrize('interp', [Interp.bilinear, Interp.cubic, Interp.lanczos])
def test_process_interp(pan_sharpen: PanSharpen, tmp_path: Path, interp: Interp):
    """Test the ``PanSharpen.process()`` ``interp`` argument by comparing with an
    ``interp='nearest'`` image.
    """
    ref_file = tmp_path.joinpath('ref.tif')
    pan_sharpen.process(ref_file, compress='deflate', interp=Interp.nearest)
    test_file = tmp_path.joinpath('test.tif')
    pan_sharpen.process(test_file, compress='deflate', interp=interp)

    with rio.open(ref_file, 'r') as ref_im, rio.open(test_file, 'r') as test_im:
        ref_array = ref_im.read(out_dtype='float32')
        test_array = test_im.read(out_dtype='float32')

    # compare
    assert np.any(test_array != ref_array)
    abs_diff = np.abs(test_array - ref_array)
    assert abs_diff.mean() < 5
    assert abs_diff.std() < 5


@pytest.mark.parametrize('driver', Driver)
def test_process_driver(pan_sharpen: PanSharpen, tmp_path: Path, driver: Driver):
    """Test the ``PanSharpen.process()`` ``driver`` argument."""
    out_file = tmp_path.joinpath('pan_sharp.tif')
    pan_sharpen.process(out_file, driver=driver)
    assert out_file.exists()

    with rio.open(out_file, 'r') as out_im:
        assert out_im.driver.lower() == 'gtiff'
        im_struct = out_im.tags(ns='IMAGE_STRUCTURE')
        if driver is Driver.gtiff:
            assert 'LAYOUT' not in im_struct or im_struct['LAYOUT'].lower() != 'cog'
        else:
            assert im_struct['LAYOUT'].lower() == 'cog'


@pytest.mark.parametrize(
    'write_mask, compress',
    [(None, Compress.jpeg), (None, Compress.deflate), (False, None), (True, None)],
)
def test_process_write_mask(
    pan_sharpen: PanSharpen, tmp_path: Path, write_mask: bool, compress: Compress
):
    """Test the ``PanSharpen.process()`` ``write_mask`` argument and its default value interaction
    with compression.
    """
    out_file = tmp_path.joinpath('pan_sharp.tif')
    pan_sharpen.process(out_file, compress=compress, write_mask=write_mask)

    if write_mask is None:
        write_mask = True if compress is Compress.jpeg else False
    mask_flag = MaskFlags.per_dataset if write_mask else MaskFlags.nodata

    with rio.open(out_file, 'r') as out_im:
        assert all([mf[0] == mask_flag for mf in out_im.mask_flag_enums])
        assert (
            (out_im.nodata is None)
            if write_mask
            else common.nan_equals(out_im.nodata, common._nodata_vals[out_im.dtypes[0]])
        )


@pytest.mark.parametrize(
    'ms_dtype, out_dtype',
    [('uint8', None), ('float32', None), *[(None, dt) for dt in common._nodata_vals.keys()]],
)
def test_process_dtype(
    pan_file: Path, ms_file: Path, tmp_path: Path, ms_dtype: str, out_dtype: str
):
    """Test the ``PanSharpen.process()`` ``dtype`` argument and its default value interaction with
    the multispectral image dtype.
    """
    with rio.open(ms_file, 'r') as ms_im_:
        profile = dict(width=ms_im_.width, height=ms_im_.height, dtype=ms_dtype)
        with WarpedVRT(ms_im_, **profile) as ms_im:
            pan_sharp = PanSharpen(pan_file, ms_im)
            out_file = tmp_path.joinpath('pan_sharp.tif')
            pan_sharp.process(out_file, dtype=out_dtype)

        out_dtype = out_dtype or ms_im.dtypes[0]

    with rio.open(out_file, 'r') as out_im:
        assert out_im.profile['dtype'] == out_dtype


@pytest.mark.parametrize(
    'compress, dtype', [(None, 'uint8'), (None, 'float32'), *[(c, None) for c in Compress]]
)
def test_process_compress(pan_sharpen: PanSharpen, tmp_path: Path, compress: Compress, dtype: str):
    """Test the ``PanSharpen.process()`` ``compress`` argument and its default value interaction
    with ``dtype``.
    """
    out_file = tmp_path.joinpath('pan_sharp.tif')
    pan_sharpen.process(out_file, dtype=dtype, compress=compress)

    if compress is None:
        compress = Compress.jpeg if dtype == 'uint8' else Compress.deflate

    with rio.open(out_file, 'r') as out_im:
        assert out_im.profile['compress'] == compress


@pytest.mark.parametrize('build_ovw, driver', [*product([True, False], Driver)])
def test_process_overview(pan_sharpen: PanSharpen, tmp_path: Path, build_ovw: bool, driver: Driver):
    """Test ``PanSharpen.process()`` creates overview(s) according to the ``build_ovw`` value."""
    out_file = tmp_path.joinpath('pan_sharp.tif')
    pan_sharpen.process(out_file, driver=driver, build_ovw=build_ovw)
    assert out_file.exists()

    with rio.open(out_file, 'r') as out_im:
        assert min(out_im.shape) >= 512
        assert len(out_im.overviews(1)) > 0 if build_ovw else len(out_im.overviews(1)) == 0


@pytest.mark.parametrize('driver', Driver)
def test_process_colorinterp(
    pan_file: Path, ms_neg_band_file: Path, tmp_path: Path, driver: Driver
):
    """Test ``PanSharpen.process()`` copies ``colorinterp`` from source to pan-sharpened image."""
    out_file = tmp_path.joinpath('pan_sharp.tif')
    pan_sharp = PanSharpen(pan_file, ms_neg_band_file)
    pan_sharp.process(out_file)
    assert out_file.exists()

    with rio.open(ms_neg_band_file, 'r') as ms_im, rio.open(out_file, 'r') as out_im:
        assert out_im.colorinterp == ms_im.colorinterp


def test_process_creation_options(pan_sharpen: PanSharpen, tmp_path: Path):
    """Test ``PanSharpen.process()`` configures the pan-sharpened image with ``creation_options``."""
    out_file = tmp_path.joinpath('pan_sharp.tif')
    pan_sharpen.process(
        out_file, creation_options=dict(tiled=True, compress='jpeg', jpeg_quality=50)
    )
    assert out_file.exists()

    with rio.open(out_file, 'r') as out_im:
        assert out_im.profile['tiled'] == True
        assert out_im.profile['compress'].lower() == 'jpeg'
        assert out_im.tags(ns='IMAGE_STRUCTURE')['JPEG_QUALITY'] == '50'


def test_process_progress(pan_sharpen: PanSharpen, tmp_path: Path, capsys: pytest.CaptureFixture):
    """Test ``PanSharpen.process()`` progress bar display."""
    # default bars
    out_file = tmp_path.joinpath('pan_sharp.tif')
    pan_sharpen.process(out_file, progress=True)
    cap = capsys.readouterr()
    progress_words = ['Statistics', 'Sharpening', 'blocks', '100%']
    assert all([w in cap.err for w in progress_words])

    # no bars
    pan_sharpen.process(out_file, overwrite=True, progress=False)
    cap = capsys.readouterr()
    assert not any([w in cap.err for w in progress_words])

    # custom bars
    descs = ['bar1', 'bar2']
    pan_sharpen.process(out_file, overwrite=True, progress=[dict(desc=d) for d in descs])
    cap = capsys.readouterr()
    assert all([w in cap.err for w in descs])



# Copyright The Orthority Contributors.
#
# This file is part of Orthority.
#
# Orthority is free software: you can redistribute it and/or modify it under the terms of the GNU
# Affero General Public License as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# Orthority is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Orthority.
# If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import numpy as np
import pytest
from rasterio.rpc import RPC

from orthority import fit
from orthority.camera import RpcCamera
from orthority.enums import RpcRefine


@pytest.mark.parametrize('shift, drift', [((5.0, 10.0), None), ((5.0, 10.0), (1.2, 0.8))])
def test_refine_rpc(
    rpc: dict, im_size: tuple[int, int], shift: tuple[float, float], drift: tuple[float, float]
):
    """Test ``refine_rpc()`` correctly refines an RPC model by testing it against refinement GCPs.
    """
    # create affine transform to realise shift & drift
    method = RpcRefine.shift if not drift else RpcRefine.shift_drift
    drift = (1.0, 1.0) if not drift else drift
    refine_tform = np.eye(2, 3)
    refine_tform[:, -1] = shift
    refine_tform[:, :2] = np.diag(drift)

    # generate GCPs randomly spread over the image, with known pixel shift/drift
    n = 10
    camera = RpcCamera(im_size, rpc)
    ji_rpc = (np.random.rand(2, n) - 0.5) * np.array([im_size]).T
    z = rpc['height_off'] / 2 + (np.random.rand(ji_rpc.shape[1]) * rpc['height_off'] / 4)
    xyz = camera.pixel_to_world_z(ji_rpc, z)

    ji_rpc_ = np.vstack((ji_rpc, np.ones(ji_rpc.shape[1])))
    ji_gcp = refine_tform.dot(ji_rpc_)
    gcps = []
    for ji_pt, xyz_pt in zip(ji_gcp.T, xyz.T):
        gcp = dict(ji=tuple(ji_pt.tolist()), xyz=tuple(xyz_pt.tolist()))
        gcps.append(gcp)

    # refine RPC with GCPs
    refined_rpc = fit.refine_rpc(rpc, gcps, method=method)

    # test refined RPC model against original GCPs
    refined_camera = RpcCamera(im_size, refined_rpc)
    xyz_test = refined_camera.pixel_to_world_z(ji_gcp, z)
    assert xyz_test == pytest.approx(xyz, abs=1e-6)


def test_refine_rpc_type(rpc: dict):
    """Test ``refine_rpc()`` works with an RPC dict or :class:`~rasterio.rpc.RPC` object."""
    gcp = dict(
        ji=(rpc['samp_off'], rpc['line_off']),
        xyz=(rpc['long_off'], rpc['lat_off'], rpc['height_off']),
    )
    fit.refine_rpc(rpc, [gcp])
    fit.refine_rpc(RPC(**rpc), [gcp])


@pytest.mark.parametrize('method, min_gcps', [(RpcRefine.shift, 1), (RpcRefine.shift_drift, 2)])
def test_refine_num_gcps(rpc: dict, im_size: tuple[int, int], method: RpcRefine, min_gcps: int):
    """Test ``refine_rpc()`` works with the minimum allowed GCPs and raises an error otherwise."""
    camera = RpcCamera(im_size, rpc)
    gcps = []
    for i in range(min_gcps):
        ji = np.array([rpc['samp_off'], rpc['line_off']]) + np.random.randn(2) * 10
        xyz = camera.pixel_to_world_z(ji.reshape(-1, 1), z=rpc['height_off']).T
        gcp = dict(ji=(*ji,), xyz=(*xyz.squeeze(),))
        gcps.append(gcp)

    fit.refine_rpc(RPC(**rpc), gcps, method=method)
    with pytest.raises(ValueError) as ex:
        fit.refine_rpc(rpc, gcps[:-1], method=method)
    assert 'At least' in str(ex.value)



# Copyright The Orthority Contributors.
#
# This file is part of Orthority.
#
# Orthority is free software: you can redistribute it and/or modify it under the terms of the GNU
# Affero General Public License as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# Orthority is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Orthority.
# If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from contextlib import contextmanager
from io import BytesIO, TextIOWrapper
from pathlib import Path

import fsspec
import numpy as np
import pytest
import rasterio as rio
from rasterio.enums import PhotometricInterp, MaskFlags

from orthority import common
from orthority.enums import Compress, Driver
from orthority.errors import OrthorityError
from tests.conftest import checkerboard, create_profile


@pytest.mark.parametrize(
    'schema, coll',
    [
        ({str: int}, dict(a=1, b=2)),
        (dict(a=1, b=2.2, c='c'), dict(a=1, b=2.2, c='c')),
        (dict(dict(a=1)), dict(dict(a=1))),
        ([int], [1, 2]),
        ([1, 2.2, 'c'], [1, 2.2, 'c']),
        ([[1]], [[1]]),
        ([dict], [dict(dict(a=1))]),
        (dict(a=None, b=None), dict(a=1, b=2)),
        ([None], [1, 2]),
    ],
)
def test_validate_collection(schema: list | dict, coll: list | dict):
    """Test ``validate_collection`` passes valid ``schema`` / ``coll`` combinations."""
    common.validate_collection(schema, coll)


@pytest.mark.parametrize(
    'schema, coll',
    [
        ({str: int}, {1: 1}),
        ({str: int}, {'a': 'a'}),
        (dict(a=1), dict(a=2)),
        (dict(a=1), dict(b=1)),
        ([int], ['a']),
        ([1, 2.2], [1, 2.3]),
        ([dict(dict(a=1))], [dict()]),
        (dict(a=[]), dict()),
        ([[]], [dict()]),
        ([dict()], [[]]),
    ],
)
def test_validate_collection_error(schema: list | dict, coll: list | dict):
    """Test ``validate_collection`` fails invalid ``schema`` / ``coll`` combinations."""
    with pytest.raises((ValueError, TypeError, KeyError)):
        common.validate_collection(schema, coll)


@pytest.mark.parametrize('file', ['odm_image_file', 'odm_image_url'])
def test_get_filename(file: str, request: pytest.FixtureRequest):
    """Test get_filename() returns with different ``file`` objects."""
    file = request.getfixturevalue(file)
    exp_val = Path(file).name
    assert common.get_filename(file) == exp_val
    ofile = fsspec.open(file, 'rb')
    assert common.get_filename(ofile) == exp_val
    with common.Open(file, 'rb') as f:
        assert common.get_filename(f) == exp_val
    with common.OpenRaster(file, 'r') as f:
        assert common.get_filename(f) == exp_val


@pytest.mark.parametrize('file', ['odm_image_file', 'odm_image_url'])
def test_join_ofile(file: str, request: pytest.FixtureRequest):
    """Test join_ofile() returns valid OpenFile instances for existing paths / URIs,
    with different base path / URI types, and with and without trailing slashes on the base path
    / URI.
    """
    file = request.getfixturevalue(file)
    ofile = fsspec.open(file)
    parts = str(file).replace('\\', '/').split('/')

    path_types = [str, fsspec.open, Path] if isinstance(file, Path) else [str, fsspec.open]
    for path_type in path_types:
        for pidx in [1, 2]:
            _base_path = '/'.join(parts[:-pidx])
            rel_path = '/'.join(parts[-pidx:])

            for base_path in map(path_type, [_base_path, _base_path + '/']):
                join_ofile = common.join_ofile(base_path, rel_path)
                assert isinstance(join_ofile, fsspec.core.OpenFile)
                assert join_ofile.path == ofile.path
                assert join_ofile.fs.exists(join_ofile.path)


@pytest.mark.parametrize('raster_file', ['odm_image_file', 'odm_image_url'])
def test_open_raster_read(raster_file: str, request: pytest.FixtureRequest):
    """Test OpenRaster in ``'r'`` mode with path / URI, fsspec OpenFile and dataset objects."""
    raster_file = request.getfixturevalue(raster_file)

    def _test_open_raster_read(_raster_file, test_closed: bool = True):
        """Test OpenRaster for the given ``_raster_file`` object."""
        with common.OpenRaster(_raster_file, 'r') as _ds:
            assert isinstance(_ds, rio.DatasetReader)
            assert not _ds.closed
            assert _ds.filename == Path(raster_file).name
            _ = _ds.read(1, window=_ds.block_window(1, 0, 0))
        if test_closed:
            assert _ds.closed

    _test_open_raster_read(raster_file)
    _test_open_raster_read(fsspec.open(raster_file, 'rb'))
    with common.OpenRaster(raster_file, 'r') as ds:
        _test_open_raster_read(ds, test_closed=False)
    assert ds.closed


def test_open_raster_write(tmp_path: Path):
    """Test OpenRaster in ``'w'`` mode with path / URI, fsspec OpenFile and dataset objects."""
    raster_file = tmp_path.joinpath('test_open_raster_write.tif')
    profile = rio.default_gtiff_profile
    array = np.ones((256, 256), dtype=profile['dtype'])
    profile.update(width=array.shape[1], height=array.shape[0], count=1)

    @contextmanager
    def _test_temp_file(filename: Path):
        """Test ``filename`` exists on exit, then delete."""
        try:
            yield filename
        finally:
            assert filename.exists()
            filename.unlink()

    def _test_open_raster_write(_raster_file, test_closed: bool = True):
        """Test OpenRaster for the given ``_raster_file`` object."""
        with common.OpenRaster(_raster_file, 'w', **profile) as _ds:
            assert isinstance(_ds, rio.io.DatasetWriter)
            assert not _ds.closed
            assert _ds.filename == Path(raster_file).name
            _ds.write(array, indexes=1)
        if test_closed:
            assert _ds.closed

    with _test_temp_file(raster_file):
        _test_open_raster_write(raster_file)
    with _test_temp_file(raster_file):
        _test_open_raster_write(fsspec.open(str(raster_file), 'wb'))
    with _test_temp_file(raster_file):
        with common.OpenRaster(raster_file, 'w', **profile) as ds:
            _test_open_raster_write(ds, test_closed=False)
        assert ds.closed


def test_open_raster_overwrite(tmp_path: Path):
    """Test the OpenRaster ``overwrite`` argument."""
    raster_file = tmp_path.joinpath('test_open_raster_overwrite.tif')
    profile = rio.default_gtiff_profile
    array = np.ones((256, 256), dtype=profile['dtype'])
    profile.update(width=array.shape[1], height=array.shape[0], count=1)
    raster_file.touch()

    def _test_open_raster_overwrite(_raster_file, overwrite: bool):
        with common.OpenRaster(_raster_file, 'w', overwrite=overwrite, **profile) as _ds:
            _ds.write(array, indexes=1)

    # test overwriting an existing file with overwrite=True
    _test_open_raster_overwrite(raster_file, overwrite=True)
    _test_open_raster_overwrite(fsspec.open(str(raster_file), 'wb'), overwrite=True)

    # test writing to an existing file with overwrite=False raises FileExistsError
    with pytest.raises(FileExistsError) as ex:
        _test_open_raster_overwrite(raster_file, overwrite=False)
    assert raster_file.name in str(ex.value)
    with pytest.raises(FileExistsError) as ex:
        _test_open_raster_overwrite(fsspec.open(str(raster_file), 'wb'), overwrite=False)
    assert raster_file.name in str(ex.value)


@pytest.mark.parametrize(
    'raster_file',
    [
        'unknown/unknown.tif',
        # TODO: add in URL tests when rio.open(opener=...) is working
        # 'https://un.known/unknown.tif',
        # 'https://raw.githubusercontent.com/leftfield-geospatial/orthority/main/unknown.tif',
    ],
)
def test_open_raster_not_found_error(raster_file: str):
    """Test OpenRaster raises a FileNotFoundError error with non-existing file path / URIs."""
    with pytest.raises(FileNotFoundError):
        with common.OpenRaster(raster_file, 'r'):
            pass
    ofile = fsspec.open(raster_file, 'rb')
    with pytest.raises(FileNotFoundError):
        with common.OpenRaster(ofile, 'r'):
            pass


@pytest.mark.parametrize(
    'file, kwargs',
    [
        ('ngi_oty_int_param_file', dict(newline=None, encoding='utf8')),
        ('ngi_oty_int_param_url', dict()),
    ],
)
def test_open_read(file: str, kwargs: dict, request: pytest.FixtureRequest):
    """Test Open in ``'rt'`` mode with path / URI, fsspec OpenFile and file objects."""
    file = request.getfixturevalue(file)

    def _test_open_read(_file, test_closed: bool = True, **kwargs):
        """Test Open for the given ``_file`` object."""
        with common.Open(_file, 'rt', **kwargs) as _f:
            assert isinstance(_f, TextIOWrapper)
            assert not _f.closed
            assert _f.filename == Path(file).name
            _ = _f.read()
        if test_closed:
            assert _f.closed

    _test_open_read(file)
    _test_open_read(fsspec.open(file, 'rt'))
    with common.Open(file, 'rt', **kwargs) as f:
        _test_open_read(f, test_closed=False, **kwargs)
    assert f.closed


def test_open_write(tmp_path: Path):
    """Test Open in ``'w'`` mode with path / URI, fsspec OpenFile and file objects."""
    file = tmp_path.joinpath('test_open_write.txt')

    @contextmanager
    def _test_temp_file(filename: Path):
        """Test ``filename`` exists on exit, then delete."""
        try:
            yield filename
        finally:
            assert filename.exists()
            assert filename.read_text()
            filename.unlink()

    def _test_open_write(_file, test_closed: bool = True, **kwargs):
        """Test Open for the given ``_file`` object."""
        with common.Open(_file, 'wt', **kwargs) as _f:
            assert isinstance(_f, TextIOWrapper)
            assert not _f.closed
            assert _f.filename == Path(file).name
            _f.write('test')
        if test_closed:
            assert _f.closed

    kwargs = dict(newline=None, encoding='utf8')
    with _test_temp_file(file):
        _test_open_write(file, **kwargs)
    with _test_temp_file(file):
        _test_open_write(fsspec.open(str(file), 'wt'), **kwargs)
    with _test_temp_file(file):
        with common.Open(file, 'wt', **kwargs) as f:
            _test_open_write(f, test_closed=False, **kwargs)
        assert f.closed


def test_open_overwrite(tmp_path: Path):
    """Test the Open ``overwrite`` argument."""
    file = tmp_path.joinpath('test_open_overwrite.txt')
    file.touch()

    def _test_open_overwrite(_file, overwrite: bool):
        with common.Open(_file, 'wt', overwrite=overwrite) as _f:
            _f.write('test')

    # test overwriting an existing file with overwrite=True
    _test_open_overwrite(file, overwrite=True)
    _test_open_overwrite(fsspec.open(str(file), 'wt'), overwrite=True)

    # test writing to an existing file with overwrite=False raises FileExistsError
    with pytest.raises(FileExistsError) as ex:
        _test_open_overwrite(file, overwrite=False)
    assert file.name in str(ex.value)
    with pytest.raises(FileExistsError) as ex:
        _test_open_overwrite(fsspec.open(str(file), 'wt'), overwrite=False)
    assert file.name in str(ex.value)


@pytest.mark.parametrize(
    'file',
    ['unknown/unknown.txt', 'https://un.known/unknown.txt', 'https://github.com/unknown.txt'],
)
def test_open_not_found_error(file: str):
    """Test Open raises a FileNotFoundError error with non-existing file path / URIs."""
    with pytest.raises(FileNotFoundError):
        with common.Open(file, 'rt'):
            pass
    ofile = fsspec.open(file, 'rt')
    with pytest.raises(FileNotFoundError):
        with common.Open(ofile, 'rt'):
            pass


@pytest.mark.parametrize('driver', Driver)
def test_create_profile_driver(driver: Driver):
    """Test ``create_profile()`` sets ``driver``."""
    profile, write_mask = common.create_profile(driver=driver, shape=(1, 1, 1), dtype='uint8')
    assert profile['driver'] == driver


def test_create_profile_shape():
    """Test ``create_profile()`` sets ``shape`` related items."""
    shape = (3, 2, 1)
    profile, write_mask = common.create_profile(driver=Driver.gtiff, shape=shape, dtype='uint8')
    assert profile['count'] == shape[0]
    assert profile['height'] == shape[1]
    assert profile['width'] == shape[2]


def test_create_profile_dtype():
    """Test ``create_profile()`` sets ``dtype``."""
    dtype = 'uint16'
    profile, write_mask = common.create_profile(driver=Driver.gtiff, shape=(1, 1, 1), dtype=dtype)
    assert profile['dtype'] == dtype


@pytest.mark.parametrize(
    'driver, exp_profile',
    [
        (Driver.gtiff, dict(tiled=True, blockxsize=512, blockysize=512, bigtiff='if_safer')),
        (Driver.cog, dict(blocksize=512, bigtiff='if_safer')),
    ],
)
def test_create_profile_non_config_items(driver: Driver, exp_profile: dict):
    """Test ``create_profile()`` non-configurable items."""
    profile, write_mask = common.create_profile(driver=driver, shape=(1, 1, 1), dtype='uint8')
    for k, v in exp_profile.items():
        assert profile[k] == v


@pytest.mark.parametrize(
    'dtype, compress, exp_value',
    [
        # compress defaults to 'jpeg' if dtype is uint8, and to 'deflate' otherwise
        ('uint8', None, 'jpeg'),
        ('uint16', None, 'deflate'),
        # compress is copied through as is when provided
        ('uint8', 'deflate', 'deflate'),
        ('uint16', 'jpeg', 'jpeg'),
        ('int16', 'lzw', 'lzw'),
    ],
)
def test_create_profile_compress(dtype: str, compress: str, exp_value: str):
    """Test ``create_profile()`` ``compress`` configuration."""
    for driver in Driver:
        profile, write_mask = common.create_profile(
            driver=driver, shape=(1, 1, 1), dtype=dtype, compress=compress
        )

        assert profile['compress'] == exp_value


@pytest.mark.parametrize(
    'compress, shape, exp_value',
    [
        # jpeg compression with 3 bands should give ''ycbcr'
        ('jpeg', (3, 1, 1), 'ycbcr'),
        # any other configuration should give None
        ('deflate', (3, 1, 1), None),
    ],
)
def test_create_profile_photometric(compress: str, shape: tuple, exp_value: str | None):
    """Test ``create_profile()`` ``photometric`` configuration with ``driver='gtiff'``."""
    profile, write_mask = common.create_profile(
        driver=Driver.gtiff, shape=shape, dtype='uint8', compress=compress
    )

    assert profile['dtype'] == 'uint8'
    assert profile['compress'] == compress
    assert profile.get('photometric', None) == exp_value


@pytest.mark.parametrize('driver', Driver)
def test_create_profile_12bit_jpeg(driver: Driver):
    """Test ``create_profile()`` correctly configures a 12bit jpeg profile."""
    # Note: depending on how rasterio is built, it may or may not support reading/writing 12 bit
    # jpeg compression.  This test just checks the profile is correct.
    profile, write_mask = common.create_profile(
        driver=driver, shape=(1, 1, 1), dtype='uint16', compress=Compress.jpeg
    )

    assert write_mask
    assert profile['dtype'] == 'uint16'
    assert profile['compress'] == 'jpeg'
    assert 'nbits' in profile and profile['nbits'] == 12


@pytest.mark.parametrize(
    'dtype, write_mask, exp_values',
    [
        # with a uint8 dtype (i.e. jpeg compression) write_mask should default to True and nodata
        # to None
        ('uint8', None, (True, None)),
        # with other dtypes write_mask should default to False and nodata to the dtype min value /
        # nan
        ('uint16', None, (False, 0)),
        ('int16', None, (False, np.iinfo('int16').min)),
        # when write_mask is False, it should be copied through as is, and nodata set to the
        # dtype min value / nan
        ('uint8', False, (False, 0)),
        ('float32', False, (False, float('nan'))),
        ('float64', False, (False, float('nan'))),
        # when write_mask is True, it should be copied through as is, and nodata set to None
        ('uint8', True, (True, None)),
    ],
)
def test_create_profile_write_mask_nodata(dtype: str, write_mask: bool | None, exp_values: tuple):
    """Test ``create_profile()`` correctly sets ``write_mask`` and ``nodata``."""
    for driver in Driver:
        profile, write_mask = common.create_profile(
            driver=driver, shape=(1, 1, 1), dtype=dtype, write_mask=write_mask
        )

        assert write_mask is exp_values[0]
        if profile['nodata'] is None or exp_values[1] is None:
            assert profile['nodata'] is None and exp_values[1] is None
        else:
            assert common.nan_equals(profile['nodata'], exp_values[1])


def test_create_profile_driver_error():
    """Test ``create_profile()`` raises an error when ``driver`` is not supported."""
    with pytest.raises(ValueError) as ex:
        common.create_profile(driver='other', shape=(1, 1, 1), dtype='uint8')

    assert 'other' in str(ex.value)


def test_create_profile_dtype_error():
    """Test ``create_profile()`` raises an error when ``dtype`` is not supported."""
    with pytest.raises(OrthorityError) as ex:
        common.create_profile(
            driver=Driver.gtiff, shape=(1, 1, 1), dtype='uint32', compress=Compress.jpeg
        )

    assert 'uint32' in str(ex.value)


@pytest.mark.parametrize('driver', Driver)
def test_create_profile_compress_error(driver: Driver):
    """Test ``create_profile()`` raises an error when ``compress`` is JPEG and ``dtype`` is not
    uint8 or uint16.
    """
    with pytest.raises(OrthorityError) as ex:
        common.create_profile(
            driver=driver, shape=(1, 1, 1), dtype='float32', compress=Compress.jpeg
        )

    assert 'uint8' in str(ex.value)


def test_create_profile_creation_options():
    """Test ``create_profile()`` populates the profile with ``creation_options`` creation options
    only.
    """
    driver = Driver.gtiff
    dtype = 'uint16'
    shape = (3, 2, 1)
    creation_options = dict(item1=123, item2='abc')
    profile, write_mask = common.create_profile(
        driver=driver, shape=shape, dtype=dtype, creation_options=creation_options
    )

    exp_profile = dict(
        driver=driver.value,
        dtype=dtype,
        width=shape[2],
        height=shape[1],
        count=shape[0],
        bigtiff='if_safer',
        nodata=0,
        **creation_options,
    )
    assert write_mask == False
    assert profile == exp_profile


@pytest.mark.parametrize(
    'driver, shape, dtype, compress, write_mask',
    [
        (Driver.gtiff, (1, 1, 1), 'uint8', Compress.jpeg, False),
        (Driver.gtiff, (1, 1, 1), 'uint16', Compress.deflate, True),
        (Driver.gtiff, (3, 1, 1), 'float32', Compress.lzw, False),
        (Driver.gtiff, (3, 1, 1), 'uint8', Compress.jpeg, True),
        (Driver.cog, (1, 1, 1), 'uint8', Compress.jpeg, False),
        (Driver.cog, (1, 1, 1), 'uint16', Compress.deflate, True),
        (Driver.cog, (3, 1, 1), 'float32', Compress.lzw, False),
        (Driver.cog, (3, 1, 1), 'uint8', Compress.jpeg, True),
    ],
)
def test_create_profile_image(
    driver: Driver, shape: tuple, dtype: str, compress: Compress, write_mask: bool, tmp_path: Path
):
    """Test the ``create_profile()`` profile generates an image with the correct configuration."""
    profile, write_mask = common.create_profile(
        driver=driver, shape=shape, dtype=dtype, compress=compress, write_mask=write_mask
    )
    array = 100 * np.ones(shape, dtype=dtype)
    test_file = tmp_path.joinpath('test.tif')
    with rio.open(test_file, 'w', **profile) as im:
        im.write(array)
        if write_mask:
            im.write_mask(array > 0)

    with rio.open(test_file, 'r') as im:
        # driver
        assert im.driver.lower() == 'gtiff'
        im_struct = im.tags(ns='IMAGE_STRUCTURE')
        if driver is Driver.gtiff:
            assert 'LAYOUT' not in im_struct or im_struct['LAYOUT'].lower() != 'cog'
        else:
            assert im_struct['LAYOUT'].lower() == 'cog'

        # tiling
        assert im.profile['tiled'] == True
        assert im.profile['blockxsize'] == im.profile['blockysize'] == 512

        # dtype and compression
        assert im.dtypes[0] == dtype
        assert im.compression.value.lower() == compress

        # photometric
        if shape[0] == 3 and compress is Compress.jpeg:
            assert im.photometric is PhotometricInterp.ycbcr
        else:
            assert im.photometric is None

        # write_mask and nodata
        mask_flag = MaskFlags.per_dataset if write_mask else MaskFlags.nodata
        assert all([mf[0] == mask_flag for mf in im.mask_flag_enums])
        assert (
            (im.nodata is None)
            if write_mask
            else common.nan_equals(im.nodata, common._nodata_vals[dtype])
        )


@pytest.mark.parametrize(
    'driver, creation_options',
    [
        (
            Driver.gtiff,
            dict(
                tiled=True,
                blockxsize=256,
                blockysize=256,
                compress='jpeg',
                photometric='ycbcr',
                jpeg_quality=90,
            ),
        ),
        (Driver.cog, dict(blocksize=256, compress='jpeg', quality=90)),
    ],
)
def test_create_profile_image_creation_options(
    tmp_path: Path, driver: Driver, creation_options: dict
):
    """Test the ``create_profile()`` profile with ``creation_options`` generates an image with the
    correct configuration.
    """
    shape = (3, 1, 1)
    dtype = 'uint8'
    write_mask = True
    profile, write_mask = common.create_profile(
        driver=driver,
        shape=shape,
        dtype=dtype,
        write_mask=write_mask,
        creation_options=creation_options,
    )
    array = 100 * np.ones(shape, dtype=dtype)
    test_file = tmp_path.joinpath('test.tif')
    with rio.open(test_file, 'w', **profile) as im:
        im.write(array)
        if write_mask:
            im.write_mask(array > 0)

    with rio.open(test_file, 'r') as im:
        # driver
        assert im.driver.lower() == 'gtiff'
        im_struct = im.tags(ns='IMAGE_STRUCTURE')
        if driver is Driver.gtiff:
            assert 'LAYOUT' not in im_struct or im_struct['LAYOUT'].lower() != 'cog'
        else:
            assert im_struct['LAYOUT'].lower() == 'cog'

        # tiling
        assert im.profile['tiled'] == True
        assert im.profile['blockxsize'] == im.profile['blockysize'] == 256

        # compression
        assert im.compression.value.lower() == 'jpeg'
        assert im.photometric is PhotometricInterp.ycbcr
        assert im_struct['JPEG_QUALITY'] == '90'


@pytest.mark.parametrize(
    'src_dtype, dst_dtype',
    [
        ('uint16', 'uint8'),
        ('uint16', 'int16'),
        ('uint16', 'float32'),
        ('float32', 'uint8'),
        ('float32', 'uint16'),
        ('float32', 'int16'),
        ('float32', 'int32'),
        ('float32', 'float32'),
        ('float32', 'float64'),
        ('float64', 'float32'),
        ('float64', 'float64'),
    ],
)
def test_convert_array_dtype(src_dtype: str, dst_dtype: str):
    """Test convert_array_dtype() conversion with combinations covering rounding and clipping
    (with and w/o type promotion).
    """
    src_info = np.iinfo(src_dtype) if np.issubdtype(src_dtype, np.integer) else np.finfo(src_dtype)
    dst_info = np.iinfo(dst_dtype) if np.issubdtype(dst_dtype, np.integer) else np.finfo(dst_dtype)

    # create array that spans the src_dtype range & includes decimals
    array = np.geomspace(1, src_info.max, 50, dtype=src_dtype).reshape(5, 10)
    if src_info.min != 0:
        array = np.concatenate(
            (array, np.geomspace(-1, src_info.min, 50, dtype=src_dtype).reshape(5, 10))
        )

    # convert to dtype
    test_array = common.convert_array_dtype(array, dst_dtype)

    # create rounded & clipped array to test against
    ref_array = array
    if np.issubdtype(dst_dtype, np.integer):
        # promote dtype to clip correctly
        ref_array = ref_array.astype(np.promote_types(array.dtype, dst_dtype))
        ref_array = np.clip(np.round(ref_array), dst_info.min, dst_info.max)
    elif np.issubdtype(src_dtype, np.floating):
        # don't clip float but set out of range vals to +-inf (as np.astype does)
        ref_array[ref_array < dst_info.min] = float('-inf')
        ref_array[ref_array > dst_info.max] = float('inf')
        assert np.any(ref_array % 1 != 0)  # check contains decimals

    assert test_array.dtype == dst_dtype
    # use approx test for case of (expected) precision loss e.g. float64->float32
    assert test_array == pytest.approx(ref_array, rel=1e-6)


def test_build_overviews():
    """Test build_overviews() builds overviews for an open in-memory dataset."""
    # create in-memory dataset
    array = checkerboard((768, 1024)).astype('uint8')
    array = np.stack((array,) * 3, axis=0)
    profile = create_profile(array)
    buf = BytesIO()
    with rio.open(buf, 'w', driver='GTiff', **profile) as im:
        im.write(array)
        # build overviews
        common.build_overviews(im, min_level_pixels=256)

    with rio.open(buf, 'r') as im:
        assert len(im.overviews(1)) > 0


def test_block_windows(ms_float_src_file):
    """Test ``block_windows()`` against the corresponding Rasterio method."""
    with rio.open(ms_float_src_file, 'r') as im:
        test_block_wins = [*common.block_windows(im)]
        ref_block_wins = [win for _, win in im.block_windows(1)]

    assert test_block_wins == ref_block_wins



# Copyright The Orthority Contributors.
#
# This file is part of Orthority.
#
# Orthority is free software: you can redistribute it and/or modify it under the terms of the GNU
# Affero General Public License as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# Orthority is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Orthority.
# If not, see <https://www.gnu.org/licenses/>.

"""Factories for creating camera models from parameter files and dictionaries."""
from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from os import PathLike
from pathlib import Path
from typing import IO, Sequence

import rasterio as rio
from fsspec.core import OpenFile

from orthority import param_io, common
from orthority.camera import Camera, create_camera, FrameCamera, RpcCamera
from orthority.errors import CrsMissingError, OrthorityWarning, ParamError
from orthority.fit import refine_rpc


class Cameras(ABC):
    """Base camera factory class."""

    @abstractmethod
    def __init__(self, **kwargs):
        pass

    @property
    @abstractmethod
    def filenames(self) -> set[str]:
        """Set of filenames that have cameras."""
        pass

    @abstractmethod
    def get(self, filename: str | PathLike | OpenFile | rio.DatasetReader) -> Camera:
        """
        Return a camera object for the given image filename.

        :param filename:
            Image filename.  Can be a path, URI string, :class:`~fsspec.core.OpenFile` instance
            or dataset.

        :return:
            Camera object.
        """
        pass

    @abstractmethod
    def write_param(self, out_dir: str | PathLike | OpenFile, overwrite: bool = False):
        """
        Write camera parameters to Orthority format file(s).

        :param out_dir:
            Directory to write into.  Can be a path, URI string, or an
            :class:`~fsspec.core.OpenFile` object.
        :param overwrite:
            Whether to overwrite file(s) if they exist.
        """
        pass


class FrameCameras(Cameras):
    """
    Frame camera factory.

    :param int_param:
        Interior parameter file or dictionary.  If a file, can be a path or URI string,
        an :class:`~fsspec.core.OpenFile` object or a file object, opened in text mode (``'rt'``).
    :param ext_param:
        Exterior parameter file or dictionary.  If a file, can be a path or URI string,
        an :class:`~fsspec.core.OpenFile` object or a file object, opened in text mode (``'rt'``).
    :param io_kwargs:
        Optional dictionary of keyword arguments for the :class:`~orthority.param_io.FrameReader`
        sub-class corresponding to the exterior parameter file format. Should exclude
        ``ext_param`` or ``int_param`` file names, which are passed internally.  If ``ext_param``
        is a dictionary, these arguments are not passed to the
        :class:`~orthority.param_io.FrameReader` sub-class, but :attr:`FrameCameras.crs` is set
        with the value of a ``crs`` argument.
    :param cam_kwargs:
        Optional dictionary of keyword arguments for the
        :class:`~orthority.camera.FrameCamera` class. Should exclude interior and exterior
        parameters which are passed internally.
    """

    def __init__(
        self,
        int_param: str | PathLike | OpenFile | IO[str] | dict[str, dict],
        ext_param: str | PathLike | OpenFile | IO[str] | dict[str, dict],
        io_kwargs: dict = None,
        cam_kwargs: dict = None,
    ):
        self._int_param_dict, self._ext_param_dict, self._crs = self._read_param(
            int_param, ext_param, **(io_kwargs or {})
        )
        self._cam_kwargs = cam_kwargs or {}
        self._cameras = {}

    @classmethod
    def from_images(
        cls,
        files: Sequence[str | PathLike | OpenFile | rio.DatasetReader],
        io_kwargs: dict = None,
        cam_kwargs: dict = None,
    ):
        """
        Create frame camera factory from :doc:`image file(s) with EXIF / XMP tags
        <../file_formats/exif_xmp>`.

        :param files:
            Image file(s) to read as a list of paths or URI strings, :class:`~fsspec.core.OpenFile`
            objects in binary mode (``'rb'``), or dataset readers.
        :param io_kwargs:
            Optional dictionary of keyword arguments for the
            :class:`~orthority.param_io.ExifReader` class.  Should exclude ``files`` which is
            passed internally.
        :param cam_kwargs:
            Optional dictionary of keyword arguments for the
            :class:`~orthority.camera.FrameCamera` class. Should exclude interior and exterior
            parameters which are passed internally.
        """
        io_kwargs = io_kwargs or {}
        reader = param_io.ExifReader(files, **io_kwargs)
        int_param_dict = reader.read_int_param()
        ext_param_dict = reader.read_ext_param()
        io_kwargs.update(crs=reader.crs)
        return cls(int_param_dict, ext_param_dict, io_kwargs=io_kwargs, cam_kwargs=cam_kwargs)

    @property
    def crs(self) -> rio.CRS | None:
        """CRS of the world coordinate system."""
        return self._crs

    @property
    def filenames(self) -> set[str]:
        return set(self._ext_param_dict.keys())

    @staticmethod
    def _read_param(
        int_param: str | PathLike | OpenFile | IO[str] | dict[str, dict],
        ext_param: str | PathLike | OpenFile | IO[str] | dict[str, dict],
        **kwargs,
    ) -> tuple[dict[str, dict], dict[str, dict], rio.CRS | None]:
        """Read interior parameters, exterior parameters and world CRS."""
        crs = None
        if not isinstance(int_param, dict):
            # read interior params
            int_param_suffix = Path(common.get_filename(int_param)).suffix.lower()
            int_param_dict = None
            if int_param_suffix in ['.yaml', '.yml']:
                int_param_dict = param_io.read_oty_int_param(int_param)
            elif int_param_suffix == '.json':
                # Only read OSfM interior params here if the interior and exterior param file
                # objects are different, otherwise they are read with exteriors below. (Note that
                # int_param != ext_param is True if OpenFile / file objects are different but point
                # to the same file.)
                if int_param != ext_param:
                    int_param_dict = param_io.read_osfm_int_param(int_param)
            else:
                raise ParamError(
                    f"'{int_param_suffix}' interior parameter file type not supported."
                )
        else:
            int_param_dict = int_param

        if not isinstance(ext_param, dict):
            # read exterior params and CRS
            ext_param_suffix = Path(common.get_filename(ext_param)).suffix.lower()
            if ext_param_suffix in ['.csv', '.txt']:
                reader = param_io.CsvReader(ext_param, **kwargs)
            elif ext_param_suffix == '.json':
                reader = param_io.OsfmReader(ext_param, **kwargs)
            elif ext_param_suffix == '.geojson':
                reader = param_io.OtyReader(ext_param)
            else:
                raise ParamError(
                    f"'{ext_param_suffix}' exterior parameter file type not supported."
                )
            ext_param_dict = reader.read_ext_param()
            crs = reader.crs

            # read interior params if not read already
            int_param_dict = int_param_dict or reader.read_int_param()
        else:
            ext_param_dict = ext_param

        # copy crs from kwargs if it is not set already
        crs = crs or kwargs.get('crs', None)
        crs = rio.CRS.from_string(crs) if isinstance(crs, str) else crs

        return int_param_dict, ext_param_dict, crs

    def get(self, filename: str | PathLike | OpenFile | rio.DatasetReader) -> FrameCamera:
        # get exterior params for filename
        filename = Path(common.get_filename(filename))
        ext_param = self._ext_param_dict.get(
            filename.name, self._ext_param_dict.get(filename.stem, None)
        )
        if not ext_param:
            raise ParamError(f"Could not find exterior parameters for '{filename.name}'.")

        # get interior params for ext_param
        cam_id = ext_param.get('camera', None)
        if cam_id:
            if cam_id not in self._int_param_dict:
                raise ParamError(f"Could not find interior parameters for camera '{cam_id}'.")
            int_param = self._int_param_dict[cam_id]
        elif len(self._int_param_dict) == 1:
            int_param = list(self._int_param_dict.values())[0]
        else:
            raise ParamError(
                f"Exterior parameters for '{filename.name}' should define a 'camera' ID."
            )

        # create camera on first use
        if cam_id not in self._cameras:
            self._cameras[cam_id] = create_camera(**int_param, **self._cam_kwargs)

        # update exterior params
        self._cameras[cam_id].update(xyz=ext_param['xyz'], opk=ext_param['opk'])

        return self._cameras[cam_id]

    def write_param(self, out_dir: str | PathLike | OpenFile, overwrite: bool = False):
        # write interior params
        int_param_file = common.join_ofile(out_dir, 'int_param.yaml', mode='wt')
        param_io.write_int_param(int_param_file, self._int_param_dict, overwrite=overwrite)

        if not self.crs:
            raise CrsMissingError("A world 'crs' is required to write exterior parameters.")

        # write exterior params
        ext_param_file = common.join_ofile(out_dir, 'ext_param.geojson', mode='wt')
        param_io.write_ext_param(
            ext_param_file, self._ext_param_dict, overwrite=overwrite, crs=self.crs
        )


class RpcCameras(Cameras):
    """
    RPC camera factory.

    :param rpc_param:
        :doc:`Orthority RPC parameter file <../file_formats/oty_rpc>` or dictionary.  If a file,
        can be a path or URI string, an :class:`~fsspec.core.OpenFile` object or a file object,
        opened in text mode (``'rt'``).
    :param cam_kwargs:
        Optional dictionary of keyword arguments for the :class:`~orthority.camera.RpcCamera`
        class.  Should exclude ``im_size`` and ``rpc`` which are passed internally.
    """

    def __init__(
        self,
        rpc_param: str | PathLike | OpenFile | IO[str] | dict[str, dict],
        cam_kwargs: dict = None,
    ):
        if not isinstance(rpc_param, dict):
            self._rpc_param_dict = param_io.read_oty_rpc_param(rpc_param)
        else:
            self._rpc_param_dict = rpc_param

        self._cam_kwargs = cam_kwargs or {}
        self._cameras = {}
        self._gcp_dict = None

    @classmethod
    def from_images(
        cls,
        files: Sequence[str | PathLike | OpenFile | rio.DatasetReader],
        io_kwargs: dict = None,
        cam_kwargs: dict = None,
    ):
        """
        Create RPC camera factory from :doc:`image file(s) with RPC tags / sidecar file(s)
        <../file_formats/image_rpc>`.

        :param files:
            Image file(s) to read as a list of paths or URI strings, :class:`~fsspec.core.OpenFile`
            objects in binary mode (``'rb'``), or dataset readers.
        :param io_kwargs:
            Optional dictionary of additional arguments for
            :class:`~orthority.param_io.read_im_rpc_param`.  Should exclude ``files`` which is
            passed internally.
        :param cam_kwargs:
            Optional dictionary of keyword arguments for the :class:`~orthority.camera.RpcCamera`
            class.  Should exclude ``im_size`` and ``rpc`` which are passed internally.
        """
        io_kwargs = io_kwargs or {}
        rpc_param_dict = param_io.read_im_rpc_param(files, **io_kwargs)
        return cls(rpc_param_dict, cam_kwargs=cam_kwargs)

    @property
    def filenames(self) -> set[str]:
        return set(self._rpc_param_dict.keys())

    def refine(
        self,
        gcps: (
            str
            | PathLike
            | OpenFile
            | IO[str]
            | Sequence[str | PathLike | OpenFile | rio.DatasetReader]
            | dict[str, list[dict]]
        ),
        io_kwargs: dict = None,
        fit_kwargs: dict = None,
    ):
        """
        Refine RPC models with GCPs.

        :param gcps:
            GCPs as one of:

            - :doc:`Orthority GCP file <../file_formats/oty_gcps>` as a path or URI string,
              an :class:`~fsspec.core.OpenFile` object or file object, opened in text mode
              (``'rt'``).
            - :doc:`Image file(s) with GCP tags <../file_formats/image_gcps>` as a list of paths or
              URI strings, :class:`~fsspec.core.OpenFile` objects in binary mode (``'rb'``),
              or dataset readers.
            - GCP dictionary.
        :param io_kwargs:
            Optional dictionary of keyword arguments for
            :class:`~orthority.param_io.read_im_gcps` if ``gcps`` is a list of image file(s).
            Should exclude ``files`` which is passed internally.
        :param fit_kwargs:
            Optional dictionary of keyword arguments for :meth:`~orthority.fit.refine_rpc`.
            Should exclude ``rpc`` and ``gcps``, which are passed internally.
        """
        if gcps and not isinstance(gcps, dict):
            # read GCPs
            if isinstance(gcps, (list, tuple)):
                self._gcp_dict = param_io.read_im_gcps(gcps, **(io_kwargs or {}))
            else:
                self._gcp_dict = param_io.read_oty_gcps(gcps)
        else:
            self._gcp_dict = gcps

        # refine RPC parameters with GCPs
        for filename, rpc_param in self._rpc_param_dict.items():
            filename = Path(filename)
            gcps = self._gcp_dict.get(filename.name, self._gcp_dict.get(filename.stem, None))

            if gcps:
                self._cameras.pop(filename.name, None)  # force camera recreation
                rpc_param['rpc'] = refine_rpc(rpc_param['rpc'], gcps, **(fit_kwargs or {}))
            else:
                warnings.warn(
                    f"Could not find any GCPs for '{filename}'.", category=OrthorityWarning
                )

    def get(self, filename: str | PathLike | OpenFile | rio.DatasetReader) -> RpcCamera:
        # get RPC params for filename
        filename = Path(common.get_filename(filename))
        rpc_param = self._rpc_param_dict.get(
            filename.name, self._rpc_param_dict.get(filename.stem, None)
        )
        if not rpc_param:
            raise ParamError(f"Could not find RPC parameters for '{filename.name}'.")

        if filename.name not in self._cameras:
            self._cameras[filename.name] = create_camera(**rpc_param, **self._cam_kwargs)

        return self._cameras[filename.name]

    def write_param(self, out_dir: str | PathLike | OpenFile, overwrite: bool = False):
        """
        Write camera parameters to Orthority format file(s).

        When the models have been refined, the refined models are written, together with the GCPs.

        :param out_dir:
            Directory to write into.  Can be a path, URI string, or an
            :class:`~fsspec.core.OpenFile` object.
        :param overwrite:
            Whether to overwrite file(s) if they exist.
        """
        rpc_file = common.join_ofile(out_dir, 'rpc_param.yaml', mode='wt')
        param_io.write_rpc_param(rpc_file, self._rpc_param_dict, overwrite=overwrite)
        if self._gcp_dict:
            gcp_file = common.join_ofile(out_dir, 'gcps.geojson', mode='wt')
            param_io.write_gcps(gcp_file, self._gcp_dict, overwrite=overwrite)



# Copyright The Orthority Contributors.
#
# This file is part of Orthority.
#
# Orthority is free software: you can redistribute it and/or modify it under the terms of the GNU
# Affero General Public License as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# Orthority is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Orthority.
# If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import copy
import warnings
from pathlib import Path

import pytest
import rasterio as rio

from orthority import param_io
from orthority.camera import BrownCamera, FrameCamera
from orthority.enums import CameraType, RpcRefine
from orthority.errors import CrsMissingError, ParamError
from orthority.factory import FrameCameras, RpcCameras


@pytest.mark.parametrize(
    'int_param_file, ext_param_file',
    [
        ('ngi_oty_int_param_file', 'ngi_oty_ext_param_file'),
        ('ngi_legacy_config_file', 'ngi_xyz_opk_csv_file'),
        ('odm_reconstruction_file', 'odm_reconstruction_file'),
        ('odm_reconstruction_file', 'odm_lla_rpy_csv_file'),  # test osfm interior / exterior logic
    ],
)
def test_frame_cameras_init(
    int_param_file: str, ext_param_file: str, request: pytest.FixtureRequest
):
    """Test creating a frame camera factory with different parameter files."""
    int_param_file: Path = request.getfixturevalue(int_param_file)
    ext_param_file: Path = request.getfixturevalue(ext_param_file)

    cameras = FrameCameras(int_param_file, ext_param_file)
    assert cameras.crs is not None
    assert cameras.filenames == set(cameras._ext_param_dict.keys())
    assert len(cameras._int_param_dict) > 0
    assert len(cameras._ext_param_dict) > 0


def test_frame_cameras_init_dicts(
    mult_int_param_dict: dict, mult_ext_param_dict: dict, utm34n_crs: str
):
    """Test creating a frame camera factory with parameter dictionaries."""
    io_kwargs = dict(crs=utm34n_crs)
    cameras = FrameCameras(mult_int_param_dict, mult_ext_param_dict, io_kwargs=io_kwargs)
    assert cameras._int_param_dict == mult_int_param_dict
    assert cameras._ext_param_dict == mult_ext_param_dict

    # test crs is passed through
    assert cameras.crs == rio.CRS.from_string(io_kwargs['crs'])


def test_frame_cameras_init_io_kwargs(
    ngi_legacy_config_file: Path, ngi_legacy_csv_file: Path, ngi_crs: str
):
    """Test the frame camera factory passes ``io_kwargs`` through to the parameter reader."""
    io_kwargs = dict(crs=ngi_crs)
    cameras = FrameCameras(ngi_legacy_config_file, ngi_legacy_csv_file, io_kwargs=io_kwargs)
    assert cameras.crs == rio.CRS.from_string(ngi_crs)


def test_frame_cameras_init_cam_kwargs(odm_reconstruction_file: Path, odm_image_file: Path):
    """Test the frame camera factory passes ``cam_kwargs`` through to the camera(s)."""
    cam_kwargs = dict(distort=False, alpha=0.0)
    cameras = FrameCameras(odm_reconstruction_file, odm_reconstruction_file, cam_kwargs=cam_kwargs)
    camera = cameras.get(odm_image_file)
    assert camera.distort == cam_kwargs['distort']
    assert camera.alpha == cam_kwargs['alpha']


def test_frame_cameras_init_error(ngi_oty_int_param_file: Path, ngi_oty_ext_param_file: Path):
    """Test the frame camera factory raises errors with unrecognised parameter file extensions."""
    with pytest.raises(ParamError) as ex:
        FrameCameras(ngi_oty_ext_param_file, ngi_oty_ext_param_file)
    assert 'not supported' in str(ex.value)
    with pytest.raises(ParamError) as ex:
        FrameCameras(ngi_oty_int_param_file, ngi_oty_int_param_file)
    assert 'not supported' in str(ex.value)


def test_frame_cameras_from_images(odm_image_files: list[Path]):
    """Test creating a frame camera factory from images."""
    cameras = FrameCameras.from_images(odm_image_files)
    assert cameras.crs is not None
    assert cameras.filenames == {Path(fn).name for fn in odm_image_files}
    assert len(cameras._int_param_dict) > 0
    assert len(cameras._ext_param_dict) > 0


def test_frame_cameras_from_images_kwargs(odm_image_files: list[Path], ngi_crs: str):
    """Test creating a frame camera factory from images passes ``io_kwargs`` and ``cam_kwargs``
    through.
    """
    io_kwargs = dict(crs=ngi_crs)
    cam_kwargs = dict(distort=False, alpha=0.0)
    cameras = FrameCameras.from_images(odm_image_files, io_kwargs=io_kwargs, cam_kwargs=cam_kwargs)
    assert cameras.crs == rio.CRS.from_string(ngi_crs)

    camera = cameras.get(odm_image_files[0])
    assert camera.distort == cam_kwargs['distort']
    assert camera.alpha == cam_kwargs['alpha']


def test_frame_cameras_get(odm_reconstruction_file: Path, odm_image_files: list[Path]):
    """Test ``FrameCameras.get()`` for known filenames."""
    cameras = FrameCameras(odm_reconstruction_file, odm_reconstruction_file)
    assert cameras.filenames == {Path(fn).stem for fn in odm_image_files}
    for file in odm_image_files:
        camera = cameras.get(file)
        assert isinstance(camera, BrownCamera)

        # basic test on camera configuration
        ext_param = cameras._ext_param_dict[file.stem]
        int_param = cameras._int_param_dict[ext_param['camera']]
        assert camera.im_size == int_param['im_size']
        assert camera.pos == ext_param['xyz']


def test_frame_cameras_get_mult_camera(
    mult_int_param_dict: dict, mult_ext_param_dict: dict, utm34n_crs: str
):
    """Test ``FrameCameras.get()`` with multiple camera ID interior & exterior parameters."""
    cameras = FrameCameras(mult_int_param_dict, mult_ext_param_dict)
    for file in mult_ext_param_dict.keys():
        camera = cameras.get(file)
        assert isinstance(camera, FrameCamera)


def test_frame_cameras_get_single_camera(pinhole_int_param_dict: dict, xyz: tuple, opk: tuple):
    """Test ``FrameCameras.get()`` with an exterior parameter camera ID of ``None`` and a single
    set of interior parameters.
    """
    filename = 'some_file'
    ext_param_dict = {filename: dict(xyz=xyz, opk=opk, camera=None)}
    cameras = FrameCameras(pinhole_int_param_dict, ext_param_dict)
    camera = cameras.get(filename)
    assert isinstance(camera, FrameCamera)


def test_frame_cameras_get_filename_not_found_error(
    pinhole_int_param_dict: dict, xyz: tuple, opk: tuple
):
    """Test ``FrameCameras.get()`` raises an error when there are no exterior parameters for the
    filename.
    """
    filename = 'some_file'
    ext_param_dict = {'unknown': dict(xyz=xyz, opk=opk, camera=None)}
    cameras = FrameCameras(pinhole_int_param_dict, ext_param_dict)

    with pytest.raises(ParamError) as ex:
        cameras.get(filename)
    assert filename in str(ex.value) and 'exterior parameters' in str(ex.value)


def test_frame_cameras_get_camera_not_found_error(
    pinhole_int_param_dict: dict, xyz: tuple, opk: tuple
):
    """Test ``FrameCameras.get()`` raises an error when there are no interior parameters for the
    exterior parameter camera ID.
    """
    filename = 'some_file'
    camera = 'other'
    ext_param_dict = {filename: dict(xyz=xyz, opk=opk, camera=camera)}
    cameras = FrameCameras(pinhole_int_param_dict, ext_param_dict)

    with pytest.raises(ParamError) as ex:
        cameras.get(filename)
    assert camera in str(ex.value) and 'interior parameters' in str(ex.value)


def test_frame_cameras_get_no_cam_id_error(mult_int_param_dict: dict, xyz: tuple, opk: tuple):
    """Test ``FrameCameras.get()`` raises an error when the exterior parameter camera ID is
    ``None`` and there are multiple sets of interior parameters.
    """
    filename = 'some_file'
    ext_param_dict = {filename: dict(xyz=xyz, opk=opk, camera=None)}
    cameras = FrameCameras(mult_int_param_dict, ext_param_dict)

    with pytest.raises(ParamError) as ex:
        cameras.get(filename)
    assert filename in str(ex.value) and "'camera' ID" in str(ex.value)


def test_frame_cameras_write_param(
    mult_int_param_dict: dict, mult_ext_param_dict: dict, utm34n_crs: str, tmp_path: Path
):
    """Test writing frame camera interior and exterior parameter files."""
    io_kwargs = dict(crs=utm34n_crs)
    cameras = FrameCameras(mult_int_param_dict, mult_ext_param_dict, io_kwargs=io_kwargs)

    cameras.write_param(tmp_path)
    assert tmp_path.joinpath('int_param.yaml').exists()
    assert tmp_path.joinpath('ext_param.geojson').exists()


def test_frame_cameras_write_param_overwrite(
    mult_int_param_dict: dict, mult_ext_param_dict: dict, utm34n_crs: str, tmp_path: Path
):
    """Test writing frame camera parameter files raises an error if the destination file(s) exist,
    and overwrites them when ``overwrite=True``.
    """
    io_kwargs = dict(crs=utm34n_crs)
    cameras = FrameCameras(mult_int_param_dict, mult_ext_param_dict, io_kwargs=io_kwargs)

    int_param_file = tmp_path.joinpath('int_param.yaml')
    int_param_file.touch()
    with pytest.raises(FileExistsError) as ex:
        cameras.write_param(tmp_path)
    assert int_param_file.name in str(ex.value)

    cameras.write_param(tmp_path, overwrite=True)
    assert tmp_path.joinpath('ext_param.geojson').exists()


def test_frame_cameras_write_param_crs_error(
    mult_int_param_dict: dict, mult_ext_param_dict: dict, tmp_path: Path
):
    """Test writing frame camera parameter files raises an error if there is no world CRS."""
    cameras = FrameCameras(mult_int_param_dict, mult_ext_param_dict)
    with pytest.raises(CrsMissingError):
        cameras.write_param(tmp_path)


def test_rpc_cameras_init(rpc_param_file: Path):
    """Test creating a RPC camera factory instance from a parameter file."""
    cameras = RpcCameras(rpc_param_file)
    assert cameras.filenames == set(cameras._rpc_param_dict.keys())
    assert len(cameras._rpc_param_dict) == 1


def test_rpc_cameras_init_dict(rpc_args: dict):
    """Test creating a RPC camera factory from a parameter dictionary."""
    rpc_param_dict = {'src_image.tif': dict(cam_type=CameraType.rpc, **rpc_args)}
    cameras = RpcCameras(rpc_param_dict)
    assert cameras._rpc_param_dict == rpc_param_dict


def test_rpc_cameras_init_cam_kwargs(rpc_param_file: Path, ngi_crs: str):
    """Test the RPC camera factory passes ``cam_kwargs`` through to the camera(s)."""
    cam_kwargs = dict(crs=rio.CRS.from_string(ngi_crs))
    cameras = RpcCameras(rpc_param_file, cam_kwargs=cam_kwargs)
    camera = cameras.get(list(cameras.filenames)[0])
    assert camera.crs == cam_kwargs['crs']


def test_rpc_cameras_from_images(rpc_image_file: Path):
    """Test creating a RPC camera factory from images."""
    cameras = RpcCameras.from_images((rpc_image_file,))
    assert cameras.filenames == {rpc_image_file.name}
    assert len(cameras._rpc_param_dict) == 1


def test_rpc_cameras_from_images_kwargs(
    rpc_image_file: Path, ngi_crs: str, capsys: pytest.CaptureFixture
):
    """Test creating a RPC camera factory from images passes ``io_kwargs`` and ``cam_kwargs``
    through.
    """
    desc = 'custom'
    io_kwargs = dict(progress=dict(desc=desc))
    cam_kwargs = dict(crs=rio.CRS.from_string(ngi_crs))
    cameras = RpcCameras.from_images((rpc_image_file,), io_kwargs=io_kwargs, cam_kwargs=cam_kwargs)

    cap = capsys.readouterr()
    assert desc in cap.err

    camera = cameras.get(list(cameras.filenames)[0])
    assert camera.crs == cam_kwargs['crs']


def test_rpc_cameras_get_filename_not_found_error(rpc_param_file: Path):
    """Test ``RpcCameras.get()`` raises an error with an invalid filename."""
    cameras = RpcCameras(rpc_param_file)
    filename = 'unknown.tif'
    with pytest.raises(ParamError) as ex:
        cameras.get(filename)
    assert filename in str(ex.value) and 'could not find' in str(ex.value).lower()


def test_rpc_cameras_refine(rpc_param_file: Path, rpc_image_file: Path, gcp_file: Path):
    """Test refining a RPC camera factory with various sources of GCPs."""

    def test_refine(gcp_arg):
        """Test refining with GCPs in ``gcp_arg`` changes the RPC params."""
        cameras = RpcCameras(rpc_param_file)
        ref_rpc_param = copy.deepcopy(cameras._rpc_param_dict[rpc_image_file.name]['rpc'])
        with warnings.catch_warnings():
            warnings.simplefilter('error')  # fail on warning that no GCPs found
            cameras.refine(gcp_arg)
        assert cameras._rpc_param_dict.keys() == cameras._gcp_dict.keys()
        assert cameras._rpc_param_dict != ref_rpc_param

    test_refine([rpc_image_file])
    test_refine(gcp_file)
    test_refine(param_io.read_oty_gcps(gcp_file))


def test_rpc_cameras_refine_io_kwargs(
    rpc_param_file: Path, rpc_image_file: Path, capsys: pytest.CaptureFixture
):
    """Test refining a RPC camera factory passes through ``io_kwargs`` when the GCPs are read
    from image tags.
    """
    desc = 'custom'
    cameras = RpcCameras(rpc_param_file)
    cameras.refine([rpc_image_file], io_kwargs=dict(progress=dict(desc=desc)))
    cap = capsys.readouterr()
    assert desc in cap.err


def test_rpc_cameras_refine_ref_kwargs(rpc_param_file: Path, rpc_image_file: Path):
    """Test refining a RPC camera factory passes through ``fit_kwargs`` to ``fit.refine_rpc()``."""
    cameras = RpcCameras(rpc_param_file)
    cameras.refine([rpc_image_file], fit_kwargs=dict(method=RpcRefine.shift))
    shift_rpc_param = copy.deepcopy(cameras._rpc_param_dict[rpc_image_file.name]['rpc'])

    cameras = RpcCameras(rpc_param_file)
    cameras.refine([rpc_image_file], fit_kwargs=dict(method=RpcRefine.shift_drift))
    shift_drift_rpc_param = cameras._rpc_param_dict[rpc_image_file.name]['rpc']

    assert shift_drift_rpc_param != shift_rpc_param


def test_rpc_cameras_refine_get(rpc_param_file: Path, rpc_image_file: Path):
    """Test ``RpcCameras.get()`` returns a different camera after refinement."""
    cameras = RpcCameras(rpc_param_file)
    camera = cameras.get(rpc_image_file.name)
    cameras.refine([rpc_image_file])
    refined_camera = cameras.get(rpc_image_file.name)

    assert camera != refined_camera and camera._rpc != refined_camera._rpc



# Copyright The Orthority Contributors.
#
# This file is part of Orthority.
#
# Orthority is free software: you can redistribute it and/or modify it under the terms of the GNU
# Affero General Public License as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# Orthority is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Orthority.
# If not, see <https://www.gnu.org/licenses/>.

"""Utility functions for internal use."""
from __future__ import annotations

import cProfile
import logging
import os
import posixpath
import pstats
import tracemalloc
import warnings
from contextlib import contextmanager, ExitStack
from io import IOBase
from itertools import product
from os import PathLike
from pathlib import Path
from typing import IO, Sequence, Generator

import cv2
import fsspec
import numpy as np
import rasterio as rio
from fsspec.core import OpenFile
from rasterio.enums import Resampling

try:
    from fsspec.implementations.http import HTTPFileSystem
except ImportError:
    HTTPFileSystem = type('unknown', (), {})

from fsspec.implementations.local import LocalFileSystem
from rasterio.crs import CRS
from rasterio.errors import NotGeoreferencedWarning, RasterioIOError
from rasterio.io import DatasetReaderBase, DatasetWriter
from rasterio.windows import Window

from orthority.enums import Interp, Compress, Driver
from orthority.errors import OrthorityWarning, OrthorityError

logger = logging.getLogger(__name__)

_nodata_vals = dict(
    uint8=0, uint16=0, int16=np.iinfo('int16').min, float32=float('nan'), float64=float('nan')
)
"""Nodata values for supported dtypes.  OpenCV remap doesn't support int8 or uint32, 
and only supports int32, uint64, int64 with nearest interpolation, so these dtypes are excluded.
"""

_default_out_config = dict(
    driver=Driver.gtiff, write_mask=None, dtype=None, compress=None, build_ovw=True, overwrite=False
)
"""Default configuration values for output images."""


@contextmanager
def suppress_no_georef():
    """Context manager to suppress rasterio's NotGeoreferencedWarning."""
    # TODO: warnings.catch_warnings is not thread-safe and warnings.simplefilter should rather be
    #  called once in cli.  consider what this does to API doc examples though - perhaps it can
    #  go in __init__.py.
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', category=NotGeoreferencedWarning)
        yield


def expand_window_to_grid(win: Window, expand_pixels: tuple[int, int] = (0, 0)) -> Window:
    """Expand rasterio window extents to the nearest whole numbers."""
    col_off, col_frac = np.divmod(win.col_off - expand_pixels[1], 1)
    row_off, row_frac = np.divmod(win.row_off - expand_pixels[0], 1)
    width = np.ceil(win.width + 2 * expand_pixels[1] + col_frac)
    height = np.ceil(win.height + 2 * expand_pixels[0] + row_frac)
    exp_win = Window(int(col_off), int(row_off), int(width), int(height))
    return exp_win


def nan_equals(a: np.ndarray | float, b: np.ndarray | float) -> np.ndarray:
    """Compare two numpy objects a & b, returning true where elements of both a & b are nan."""
    return (a == b) | (np.isnan(a) & np.isnan(b))


def distort_image(camera, image: np.ndarray, nodata=0, interp=Interp.nearest) -> np.ndarray:
    """Return a distorted image given a frame camera model and source image."""

    if not np.all(np.array(image.shape[-2:][::-1]) == camera.im_size):
        raise ValueError("'image' shape should be the same as the 'camera' image size.")

    # create (j, i) pixel coords for distorted image
    j_range = np.arange(0, camera.im_size[0])
    i_range = np.arange(0, camera.im_size[1])
    j_grid, i_grid = np.meshgrid(j_range, i_range, indexing='xy')
    ji = np.array((j_grid.reshape(-1), i_grid.reshape(-1)))

    # find the corresponding undistorted/ source (j, i) pixel coords
    camera_xyz = camera._pixel_to_camera(ji)
    undist_ji = camera._K_undistort.dot(camera_xyz)[:2].astype('float32')

    def distort_band(src_array: np.ndarray, dst_array: np.ndarray):
        """Distort a 2D band array."""
        cv2.remap(
            src_array,
            undist_ji[0].reshape(image.shape[-2:]),
            undist_ji[1].reshape(image.shape[-2:]),
            Interp[interp].to_cv(),
            dst=dst_array,
            borderMode=cv2.BORDER_TRANSPARENT,
        )

    dist_image = np.full(image.shape, dtype=image.dtype, fill_value=nodata)
    for bi in range(image.shape[0]):
        distort_band(image[bi], dist_image[bi])

    return dist_image


@contextmanager
def profiler():
    """Context manager for profiling in DEBUG log level."""
    if logger.getEffectiveLevel() <= logging.DEBUG:
        proc_profile = cProfile.Profile()
        tracemalloc.start()
        proc_profile.enable()

        yield

        proc_profile.disable()
        # tottime is the total time spent in the function alone. cumtime is the total time spent
        # in the function plus all functions that this function called
        proc_stats = pstats.Stats(proc_profile).sort_stats('cumtime')
        logger.debug(f'Processing times:')
        proc_stats.print_stats(20)
        current, peak = tracemalloc.get_traced_memory()
        tracemalloc.stop()
        logger.debug(
            f"Memory usage: current: {current / 10 ** 6:.1f} MB, peak: {peak / 10 ** 6:.1f} MB"
        )
    else:
        yield


def utm_crs_from_latlon(lat: float, lon: float) -> CRS:
    """Return a 2D rasterio UTM CRS for the given (lat, lon) coordinates in degrees."""
    # adapted from https://gis.stackexchange.com/questions/269518/auto-select-suitable-utm-zone-based-on-grid-intersection
    zone = int(np.floor((lon + 180) / 6) % 60) + 1
    epsg = 32600 + zone if lat >= 0 else 32700 + zone
    return CRS.from_epsg(epsg)


def validate_collection(schema: dict | list, coll: dict | list):
    """
    Validate a nested dict / list of values (``coll``) against a nested dict / list of types,
    tuples of types, and values (``schema``).

    - All items in a ``coll`` dict are validated against the first item in the corresponding
    ``schema`` dict, if it has one item with a type key.  Otherwise, ``coll`` items are validated
    against the same key ``schema`` item.
    - All items in a ``coll`` list are validated against the first item in the corresponding
    ``schema`` list, if it has one item.  Otherwise ``coll`` items are validated against
    corresponding ``schema`` items.
    - ``coll`` values are not validated against corresponding None values in ``schema``.
    """
    # adapted from https://stackoverflow.com/questions/45812387/how-to-validate-structure-or
    #  -schema-of-dictionary-in-python
    if isinstance(schema, dict) and isinstance(coll, dict):
        # schema is a dict
        first_key = [*schema][0]
        if len(schema) == 1 and isinstance(first_key, type):
            for k in coll:
                if not isinstance(k, first_key):
                    raise TypeError(f"'{k}' is not an instance of {first_key}.")
                validate_collection(schema[first_key], coll[k])
        else:
            for k in schema:
                if k in coll:
                    validate_collection(schema[k], coll[k])
                else:
                    raise KeyError(f"No key: '{k}'.")
    elif isinstance(schema, list) and isinstance(coll, list) and len(schema) and len(coll):
        # schema is a list
        if len(schema) == 1:
            for item in coll:
                validate_collection(schema[0], item)
        else:
            if len(coll) != len(schema):
                raise ValueError(f'{coll} should have {len(schema)} items.')
            for template_item, coll_item in zip(schema, coll):
                validate_collection(template_item, coll_item)
    elif isinstance(schema, type):
        # schema is a type
        if not isinstance(coll, schema):
            raise TypeError(f"'{coll}' is not an instance of {schema}.")
    elif isinstance(schema, tuple) and all([isinstance(item, type) for item in schema]):
        # schema is a tuple of types
        if not isinstance(coll, schema):
            raise TypeError(f"'{coll}' is not an instance of any of {schema}.")
    elif isinstance(schema, (str, int, float)):
        # schema is a value of a basic type
        if not coll == schema:
            raise ValueError(f"'{coll}' does not equal '{schema}'.")
    elif schema is None:
        # don't test
        pass
    else:
        # something else is wrong
        raise ValueError("Invalid collection.")


def get_filename(file: str | PathLike | OpenFile | DatasetReaderBase | IO) -> str:
    """Return a source filename for the given ``file`` object.  If ``file`` is an
    :class:`~fsspec.core.OpenFile` instance, a :class:`~rasterio.io.DatasetReaderBase` instance
    or file object, it should have a ``filename`` attribute i.e. have been created by either
    :class:`Open` or :class:`OpenRaster`.
    """
    if isinstance(file, DatasetReaderBase):
        filename = getattr(file, 'filename', Path(file.name).name)
    elif isinstance(file, OpenFile):
        filename = getattr(file, 'filename', Path(file.path).name)
    elif isinstance(file, IOBase):
        filename = getattr(file, 'filename', Path(getattr(file, 'name', '<file object>')).name)
    else:
        filename = Path(os.fspath(file)).name
    return filename


def join_ofile(base: str | PathLike | OpenFile, rel: str, mode: str = None, **kwargs) -> OpenFile:
    """Return an fsspec OpenFile whose path is a join of the ``base`` path with the ``rel`` path."""
    if not isinstance(base, OpenFile):
        base = fsspec.open(os.fspath(base), mode or 'rt')

    joined_path = posixpath.join(base.path, rel)
    return OpenFile(base.fs, joined_path, mode=mode or base.mode, **kwargs)


class OpenRaster:
    """
    Context manager for local or remote Rasterio datasets.

    :param file:
        A path, URI, :class:`~fsspec.core.OpenFile` instance, or open dataset.  If it is an open
        dataset, it is returned unaltered on entering the context, not closed on exiting the
        context, and ``mode`` and ``kwargs`` are ignored.  If is an OpenFile instance, it should
        be open in a binary mode matching ``mode``.
    :param mode:
        Mode in which the dataset is opened.  Either ``'r'`` or ``'w'``.
    :param overwrite:
        Whether to overwrite an existing file in ``'w'`` mode.  Ignored in ``'r'`` mode.
    :param kwargs:
        Keyword arguments to pass to :func:`rasterio.open`.
    """

    def __init__(
        self,
        file: str | PathLike | DatasetReaderBase | OpenFile,
        mode: str = 'r',
        overwrite: bool = False,
        **kwargs,
    ):
        if mode not in ['r', 'w']:
            raise ValueError(f"The 'mode' argument should be either 'r' or 'w', not '{mode}'.")

        self._exit_stack = ExitStack()

        if isinstance(file, DatasetReaderBase):
            if file.closed:
                raise IOError('Dataset is closed.')
            if mode not in file.mode:
                raise IOError(
                    f"Dataset mode: '{file.mode}' not compatible with the mode argument: '{mode}'."
                )
            self._dataset = file

        elif isinstance(file, (str, PathLike, OpenFile)):
            # TODO: use the opener arg to rio.open() and pin the rasterio dependency version when
            #  rasterio 1.4 is released, rather than passing an open file object.  this should
            #  allow sidecar files to read / written (test that it does).  also test that files
            #  are not buffered in memory with this option, and currently problematic fsspec
            #  protocols (e.g. github) no longer cause a seg fault.
            if isinstance(file, OpenFile):
                if mode + 'b' != file.mode:
                    raise IOError(
                        f"OpenFile object mode: '{file.mode}' should be a binary mode matching the "
                        f"mode argument: '{mode}'."
                    )
                ofile = file
            else:
                ofile = fsspec.open(os.fspath(file), mode + 'b')

            # TODO: delete sidecar files if overwriting (note that DatasetReader has a files
            #  attribute that lists associated files)
            # TODO: if a file is partially written (e.g. because of an error), rasterio fails when
            #  overwriting it
            if not overwrite and 'w' in mode and ofile.fs.exists(ofile.path):
                raise FileExistsError(f"File exists: '{ofile.path}'")

            if isinstance(ofile.fs, (LocalFileSystem, HTTPFileSystem)):
                # use GDAL internal file system
                try:
                    self._dataset = self._exit_stack.enter_context(
                        rio.open(ofile.path, mode, **kwargs)
                    )
                except RasterioIOError as ex:
                    ex_str = str(ex)
                    if 'no such file or directory' in ex_str.lower():
                        raise FileNotFoundError(ex_str)
                    else:
                        raise
            else:
                # use fsspec file object
                file_obj = self._exit_stack.enter_context(ofile)
                self._dataset = self._exit_stack.enter_context(rio.open(file_obj, mode, **kwargs))

            # store the source filename as a dataset attribute
            self._dataset.filename = get_filename(file)

        else:
            raise TypeError(f"Unsupported 'file' type: {type(file)}")

    def __enter__(self) -> rio.DatasetReader | DatasetWriter:
        if self._dataset.closed:
            raise IOError('Dataset is closed.')
        return self._dataset

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._exit_stack.__exit__(exc_type, exc_val, exc_tb)

    def close(self):
        self._exit_stack.close()


class Open:
    """
    Context manager for local or remote file IO.

    :param file:
        A path, URI, :class:`~fsspec.core.OpenFile` instance, or file object.  If it is a file
        object, it is returned unaltered on entering the context, not closed on exiting the
        context, and ``mode`` and ``kwargs`` are ignored.  If is an OpenFile instance, it should
        be opened in ``mode`` (``kwargs`` are ignored).
    :param mode:
        Mode in which the file is opened.
    :param overwrite:
        Whether to overwrite an existing file in `'`w*'`` mode.  Ignored in ``'r*'`` mode.
    :param kwargs:
        Keyword arguments to pass to :func:`fsspec.open` or ``opener`` if it is specified.
    """

    def __init__(
        self,
        file: str | PathLike | IO | OpenFile,
        mode='rt',
        overwrite: bool = False,
        **kwargs,
    ):
        # TODO: text encoding defaults to utf-8, which can't be changed from the CLI
        self._exit_stack = ExitStack()
        if isinstance(file, IOBase):
            if file.closed:
                raise IOError('File object is closed.')
            if getattr(file, 'mode', mode) != mode:
                # note: fsspec text mode file objects do not have a mode property
                raise IOError(f"File object mode should match the mode argument: '{mode}'.")
            self._file_obj = file

        elif isinstance(file, (OpenFile, str, PathLike)):
            if isinstance(file, OpenFile):
                if mode != file.mode:
                    raise IOError(
                        f"OpenFile object mode: '{file.mode}', should match the mode argument:"
                        f" '{mode}'."
                    )
                ofile = file
            else:
                ofile = fsspec.open(os.fspath(file), mode, **kwargs)

            # overwrite could be prevented with 'x' modes, but is done this way for consistency
            # with OpenRaster & rasterio which doesn't support 'x'
            if not overwrite and 'w' in mode and ofile.fs.exists(ofile.path):
                raise FileExistsError(f"File exists: '{ofile.path}'")

            self._file_obj = self._exit_stack.enter_context(ofile)
            self._file_obj.filename = get_filename(file)

        else:
            raise TypeError(f"Unsupported 'file' type: {type(file)}")

    def __enter__(self) -> IO:
        return self._file_obj

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._exit_stack.__exit__(exc_type, exc_val, exc_tb)

    def close(self):
        self._exit_stack.close()


def create_profile(
    driver: str | Driver,
    shape: Sequence[int],
    dtype: str | np.dtype,
    compress: str | Compress | None = None,
    write_mask: bool | None = None,
    creation_options: dict | None = None,
) -> tuple[dict, bool]:
    """
    Return a partial Rasterio image profile and ``write_mask`` value for an output image given its
    configuration.

    If ``compress`` is ``None`` (the default), JPEG compression is used with the 'uint8'
    ``dtype`` and DEFLATE otherwise.  If ``write_mask`` is ``None`` (the default), it is returned
    as ``True`` when JPEG compression is used, and ``False`` otherwise.  If ``creation_options``
    are supplied, no other creation options are set.
    """
    creation_options = creation_options or {}

    # check dtype support
    dtype = str(dtype)
    if dtype not in _nodata_vals:
        raise OrthorityError(f"Data type '{dtype}' is not supported.")

    # create initial profile
    driver = Driver(driver)
    profile = dict(
        driver=driver.value,
        dtype=dtype,
        width=shape[2],
        height=shape[1],
        count=shape[0],
        bigtiff='if_safer',
    )

    # set creation options
    if len(creation_options) == 0:
        # configure compression
        if compress is None:
            compress = Compress.jpeg if dtype == 'uint8' else Compress.deflate
        else:
            compress = Compress(compress)
            if compress is Compress.jpeg:
                if dtype == 'uint16':
                    warnings.warn(
                        'Attempting a 12 bit JPEG ortho configuration.  Support is rasterio build '
                        'dependent.',
                        category=OrthorityWarning,
                    )
                    profile.update(nbits=12)
                elif dtype != 'uint8':
                    raise OrthorityError(
                        f"JPEG compression is supported for 'uint8' and 'uint16' data types only."
                    )

        profile.update(compress=compress.value)

        if driver is Driver.gtiff:
            # configure photometric interpretation and tiling
            if compress == Compress.jpeg and shape[0] == 3:
                profile.update(photometric='ycbcr')
            profile.update(tiled=True, blockxsize=512, blockysize=512)
        else:
            # Configure tiling & overviews. Overviews are not created automatically, but copied
            # from any overviews built with DatasetWriter.build_overviews().  GDAL sets
            # photometric internally.
            profile.update(blocksize=512, overviews='force_use_existing')

    else:
        profile.update(**creation_options)

    # resolve auto write_mask (=None) to write masks for jpeg compression
    if write_mask is None:
        write_mask = True if compress == Compress.jpeg else False

    # set nodata to None when writing internal masks to force external tools to use mask,
    # otherwise set by dtype
    nodata = None if write_mask else _nodata_vals[dtype]
    profile.update(nodata=nodata)

    return profile, write_mask


def convert_array_dtype(array: np.ndarray, dtype: str) -> np.array:
    """Return the ``array`` converted to ``dtype``, rounding and clipping in-place when ``dtype``
    is integer.  Adapted from :meth:`homonim.raster_array.RasterArray._convert_array_dtype`.
    """
    unsafe_cast = not np.can_cast(array.dtype, dtype, casting='safe')

    # round if converting from float to integer dtype
    if unsafe_cast and np.issubdtype(array.dtype, np.floating) and np.issubdtype(dtype, np.integer):
        np.round(array, out=array)

    # clip if converting to integer dtype with smaller range than array dtype
    if unsafe_cast and np.issubdtype(dtype, np.integer):
        src_info = (
            np.iinfo(array.dtype)
            if np.issubdtype(array.dtype, np.integer)
            else np.finfo(array.dtype)
        )
        dst_info = np.iinfo(dtype)
        if src_info.min < dst_info.min or src_info.max > dst_info.max:
            # promote array dtype to be able to represent destination dtype exactly (if
            # possible) to clip correctly
            array = array.astype(np.promote_types(array.dtype, dtype))
            np.clip(array, dst_info.min, dst_info.max, out=array)

    # convert dtype (ignoring numpy warnings for float overflow or cast of nan to integer)
    with np.errstate(invalid='ignore', over='ignore'):
        array = array.astype(dtype, copy=False, casting='unsafe')

    return array


def build_overviews(
    im: DatasetWriter,
    max_num_levels: int = 8,
    min_level_pixels: int = 256,
    resampling=Resampling.average,
) -> None:
    """
    Build internal overviews for an open rasterio dataset.  Each overview level is decimated by a
    factor of 2.  The number of overview levels is determined by whichever of the
    ``max_num_levels`` or ``min_level_pixels`` limits is reached first.

    :param im:
        Rasterio dataset opened in 'r+' or 'w' mode.
    :param max_num_levels:
        Maximum number of overview levels.
    :param min_level_pixels:
        Minimum overview width / height in pixels.
    :param resampling:
        Overview resampling method.
    """
    max_ovw_levels = int(np.min(np.log2(im.shape)))
    min_level_shape_pow2 = int(np.log2(min_level_pixels))
    num_ovw_levels = np.min([max_num_levels, max_ovw_levels - min_level_shape_pow2])
    ovw_levels = [2**m for m in range(1, num_ovw_levels + 1)]
    im.build_overviews(ovw_levels, resampling=resampling)


def get_tqdm_kwargs(**kwargs) -> dict:
    """Return a dictionary of ``tqdm`` progress bar kwargs with a standard ``bar_format``."""
    return dict(
        bar_format='{l_bar}{bar}|{n_fmt}/{total_fmt} {unit} [{elapsed}<{remaining}]',
        dynamic_ncols=True,
        **kwargs,
    )


def block_windows(
    im: DatasetReaderBase | DatasetWriter, block_shape: tuple[int, int] = None
) -> Generator[Window]:
    """Block window generator for the given image, and optional block shape."""
    driver = im.driver.lower()
    block_shape = block_shape or (
        (im.profile.get('blocksize', 512),) * 2 if driver == 'cog' else im.block_shapes[0]
    )

    xrange = range(0, im.width, block_shape[1])
    yrange = range(0, im.height, block_shape[0])
    for xstart, ystart in product(xrange, yrange):
        xstop = min(xstart + block_shape[1], im.width)
        ystop = min(ystart + block_shape[0], im.height)
        yield Window(xstart, ystart, xstop - xstart, ystop - ystart)



# Copyright The Orthority Contributors.
#
# This file is part of Orthority.
#
# Orthority is free software: you can redistribute it and/or modify it under the terms of the GNU
# Affero General Public License as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# Orthority is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Orthority.
# If not, see <https://www.gnu.org/licenses/>.
"""Pan-sharpening."""
from __future__ import annotations

import logging
import os
import warnings
from concurrent.futures import ThreadPoolExecutor, as_completed
from contextlib import ExitStack
from os import PathLike
from threading import Lock
from typing import Sequence

import numpy as np
import rasterio as rio
from fsspec.core import OpenFile
from rasterio.enums import Resampling, ColorInterp
from rasterio.vrt import WarpedVRT
from rasterio.warp import transform_bounds
from rasterio.windows import intersect, Window
from rasterio.windows import transform as window_transform
from tqdm.auto import tqdm

from orthority import common
from orthority.enums import Compress, Interp, Driver
from orthority.errors import OrthorityWarning, OrthorityError

logger = logging.getLogger(__name__)


class PanSharpen:
    """
    Pan-sharpener.

    Increases the resolution of a multispectral image to that of a panchromatic image using the
    Gram-Schmidt method (https://doi.org/10.5194/isprsarchives-XL-1-W1-239-2013).

    Panchromatic and multispectral image bounds should overlap if they are georeferenced. If one
    or both of the images are not georeferenced, they are assumed to having matching bounds.

    :param pan_file:
        Panchromatic image. Can be a path or URI string, an :class:`~fsspec.core.OpenFile` object
        in binary mode (``'rb'``), or a dataset reader.
    :param ms_file:
        Multispectral image. Can be a path or URI string, an :class:`~fsspec.core.OpenFile` object
        in binary mode (``'rb'``), or a dataset reader.
    """

    # dtype & nodata for warping and transformations
    _working_dtype = 'float64'
    _working_nodata = common._nodata_vals[_working_dtype]

    # default algorithm configuration values for PanSharpen.process()
    _default_alg_config = dict(ms_indexes=None, pan_index=1, weights=None, interp=Interp.cubic)

    def __init__(
        self,
        pan_file: str | PathLike | OpenFile | rio.DatasetReader,
        ms_file: str | PathLike | OpenFile | rio.DatasetReader,
    ):
        self._pan_file = pan_file
        self._ms_file = ms_file
        self._profiles = self._get_vrt_profiles(pan_file, ms_file)
        self._pan_lock = Lock()
        self._ms_lock = Lock()
        self._out_lock = Lock()

    @staticmethod
    def _get_vrt_profiles(pan_file, ms_file) -> dict[str, dict]:
        """Return a dictionary of WarpedVRT profiles for warping pan/MS to pan and MS grids.  The
        pan and MS grids are cropped versions of the source grids that define shared bounds.
        """
        with ExitStack() as exit_stack:
            exit_stack.enter_context(common.suppress_no_georef())

            # open files
            pan_ds = exit_stack.enter_context(common.OpenRaster(pan_file))
            ms_ds = exit_stack.enter_context(common.OpenRaster(ms_file))
            # maintain source transforms separately as un-georeferenced images get custom
            # transforms
            pan_src_transform = pan_ds.transform
            ms_src_transform = ms_ds.transform

            # find shared pan & MS bounds as pan & MS windows in their own CRSs
            if ms_ds.crs is None and ms_ds.transform.is_identity:
                # MS is not georeferenced and pan may or may not be georeferenced. Assume pan &
                # MS bounds match.
                warnings.warn(
                    'Multispectral image is not georeferenced with a CRS and affine transform - '
                    'assuming its CRS and bounds match the pan image.',
                    category=OrthorityWarning,
                )
                # create a custom MS transform that matches MS to pan bounds
                ms_scale = np.array(pan_ds.shape[::-1]) / ms_ds.shape[::-1]
                ms_src_transform = pan_ds.transform * rio.Affine.scale(*ms_scale)
                # full / matching pan & MS windows
                pan_win = Window(0, 0, pan_ds.width, pan_ds.height)
                ms_win = Window(0, 0, ms_ds.width, ms_ds.height)

            elif pan_ds.crs is None and pan_ds.transform.is_identity:
                # Pan is not georeferenced and MS is georeferenced. Assume pan & MS bounds match.
                warnings.warn(
                    'Pan image is not georeferenced with a CRS and affine transform - assuming '
                    'its CRS and bounds match the multispectral image.',
                    category=OrthorityWarning,
                )
                # create a custom pan transform that matches pan to MS bounds
                pan_scale = np.array(ms_ds.shape[::-1]) / pan_ds.shape[::-1]
                pan_src_transform = ms_ds.transform * rio.Affine.scale(*pan_scale)
                # full / matching pan & MS windows
                pan_win = Window(0, 0, pan_ds.width, pan_ds.height)
                ms_win = Window(0, 0, ms_ds.width, ms_ds.height)

            else:
                # both pan and MS are georeferenced
                # find shared bounds pan window
                ms_bounds_ = transform_bounds(ms_ds.crs, pan_ds.crs, *ms_ds.bounds)
                ms_win_ = pan_ds.window(*ms_bounds_)
                if not intersect((ms_win_, pan_ds.window(*pan_ds.bounds))):
                    raise OrthorityError('Pan and multispectral bounds do not overlap.')
                pan_win = ms_win_.intersection(pan_ds.window(*pan_ds.bounds))
                pan_win = common.expand_window_to_grid(pan_win)

                # find shared bounds MS window
                pan_bounds_ = transform_bounds(pan_ds.crs, ms_ds.crs, *pan_ds.bounds)
                pan_win_ = ms_ds.window(*pan_bounds_)
                ms_win = pan_win_.intersection(ms_ds.window(*ms_ds.bounds))
                ms_win = common.expand_window_to_grid(ms_win)

            # test resolutions
            pan_res = np.abs((pan_src_transform[0], pan_src_transform[4]))
            ms_res = np.abs((ms_src_transform[0], ms_src_transform[4]))
            if np.any(pan_res > ms_res):
                raise OrthorityError(
                    f'Pan resolution: {tuple(pan_res.tolist())} exceeds multispectral resolution: '
                    f'{tuple(ms_res.tolist())}.'
                )

            # create the WarpedVRT profiles ('src_transform' items are required for
            # un-georeferenced images)
            profiles = {}
            pan_transform = window_transform(pan_win, pan_src_transform)
            ms_transform = window_transform(ms_win, ms_src_transform)
            profiles['pan_to_pan'] = dict(
                crs=pan_ds.crs,
                src_transform=pan_src_transform,
                transform=pan_transform,
                width=pan_win.width,
                height=pan_win.height,
            )
            profiles['ms_to_pan'] = profiles['pan_to_pan'].copy()
            profiles['ms_to_pan'].update(src_transform=ms_src_transform)

            profiles['ms_to_ms'] = dict(
                crs=ms_ds.crs,
                src_transform=ms_src_transform,
                transform=ms_transform,
                width=ms_win.width,
                height=ms_win.height,
            )
            profiles['pan_to_ms'] = profiles['ms_to_ms'].copy()
            profiles['pan_to_ms'].update(src_transform=pan_src_transform)

            # add common config items
            for profile in profiles.values():
                profile.update(
                    dtype=PanSharpen._working_dtype,
                    nodata=PanSharpen._working_nodata,
                    num_threads=os.cpu_count(),
                )
        return profiles

    def _validate_pan_ms_params(
        self,
        pan_im: rio.DatasetReader,
        ms_im: rio.DatasetReader,
        pan_index: int,
        ms_indexes: Sequence[int],
        weights: Sequence[float] | None,
    ) -> tuple[Sequence[int], Sequence[float] | None]:
        """Validate pan / MS indexes and weights."""
        if pan_index <= 0 or pan_index > pan_im.count:
            pan_name = common.get_filename(pan_im)
            raise OrthorityError(
                f"Pan index {pan_index} is invalid for '{pan_name}' with {pan_im.count} band(s)"
            )

        if ms_indexes is None or len(ms_indexes) == 0:
            # default to non-alpha band indexes
            ms_indexes = [
                bi + 1 for bi in range(ms_im.count) if ms_im.colorinterp[bi] != ColorInterp.alpha
            ]

        # test ms_indexes (allows user ms_indexes that are duplicates and or alpha bands)
        ms_indexes_ = np.array(ms_indexes)
        if np.any(ms_indexes_ <= 0) or np.any(ms_indexes_ > ms_im.count):
            ms_name = common.get_filename(ms_im)
            raise OrthorityError(
                f"Multispectral indexes {tuple(ms_indexes)} contain invalid values for '{ms_name}' "
                f"with {ms_im.count} band(s)"
            )

        weights = None if weights is None or len(weights) == 0 else weights
        if weights is not None:
            if len(weights) != len(ms_indexes):
                raise OrthorityError(
                    f"There should be the same number of multispectral to pan weights "
                    f"({len(weights)}) as multispectral indexes ({len(ms_indexes)})."
                )
            if np.any(np.array(weights) < 0):
                raise OrthorityError('Weight values should greater than or equal to 0.')

        return ms_indexes, weights

    def _get_stats(
        self,
        pan_im: rio.DatasetReader,
        ms_im: rio.DatasetReader,
        pan_index: int,
        ms_indexes: Sequence[int] | np.ndarray,
        progress: dict,
    ) -> tuple[np.ndarray, np.ndarray]:
        """Return pan / MS means and covariances for the given datasets. Uses the "Numerically
        Stable Parallel Computation of (Co-)Variance" method described in
        https://doi.org/10.1145/3221269.3223036 to aggregate stats over image tiles.
        """

        def get_tile_stats(
            pan_im: rio.DatasetReader,
            ms_im: rio.DatasetReader,
            indexes: Sequence[int],
            tile_win: Window,
        ) -> tuple[int, np.ndarray, np.ndarray]:
            """Return pan / MS pixel count, mean & sum of deviation products for the given
            ``tile_win``.  Fully masked tiles produce zeros for count, means & products.
            """
            # read pan and MS tiles (without masked arrays which don't release GIL)
            with self._pan_lock:
                pan_array = pan_im.read(indexes=pan_index, window=tile_win)
            with self._ms_lock:
                ms_array = ms_im.read(indexes=ms_indexes, window=tile_win)

            # mask and combine pan & MS
            pan_mask = np.logical_not(common.nan_equals(pan_array, pan_im.nodata))
            ms_mask = np.logical_not(common.nan_equals(ms_array, ms_im.nodata)).all(axis=0)
            mask = pan_mask & ms_mask
            pan_ms_array = np.concatenate(
                (pan_array[mask].reshape(1, -1), ms_array[:, mask].reshape(len(indexes), -1)),
                axis=0,
            )

            # find tile mean & sum of deviation products as in eq 12-13 in the paper
            tile_n = mask.sum()
            tile_mean = pan_ms_array.mean(axis=1) if tile_n > 0 else np.zeros(pan_ms_array.shape[0])
            pan_ms_array -= tile_mean.reshape(-1, 1)
            tile_prod = pan_ms_array.dot(pan_ms_array.T)
            return tile_n, tile_mean, tile_prod

        # accumulate pan / MS values over tiles
        with ExitStack() as ex_stack:
            # open MS & downsampled pan WarpedVRTs that lie on a MS resolution grid, and have
            # common bounds
            pan_im = ex_stack.enter_context(
                WarpedVRT(pan_im, **self._profiles['pan_to_ms'], resampling=Resampling.average)
            )
            ms_im = ex_stack.enter_context(WarpedVRT(ms_im, **self._profiles['ms_to_ms']))

            # find tile stats in a thread pool and aggregate
            n = 0
            means = np.zeros(len(ms_indexes) + 1, dtype=self._working_dtype)
            prod = np.zeros((len(ms_indexes) + 1,) * 2, dtype=self._working_dtype)
            executor = ex_stack.enter_context(ThreadPoolExecutor(max_workers=os.cpu_count()))
            futures = [
                executor.submit(get_tile_stats, pan_im, ms_im, ms_indexes, tile_win)
                for tile_win in common.block_windows(ms_im, block_shape=(1024, 1024))
            ]

            for future in tqdm(as_completed(futures), **progress, total=len(futures)):
                try:
                    tile_n, tile_mean, tile_prod = future.result()
                except Exception as ex:
                    executor.shutdown(wait=False)
                    raise RuntimeError('Could not get tile statistics.') from ex

                if tile_n > 0:
                    # eq 21 from the paper
                    mean_diffs = tile_mean - means
                    mean_diffs_ = mean_diffs.reshape(-1, 1).dot(mean_diffs.reshape(1, -1))
                    prod += tile_prod + mean_diffs_ * (n * tile_n) / (n + tile_n)
                    # eq 17 from the paper
                    n += tile_n
                    means += mean_diffs * tile_n / n

        # convert prod to unbiased covariance (as with numpy.cov default)
        cov = prod / (n - 1)
        return means, cov

    def _get_params(self, means: np.ndarray, cov: np.ndarray, weights: np.ndarray | None) -> dict:
        """Return the Gram-Schmidt pan-sharpening parameters for the given pan / MS means &
        covariances, and optional MS to pan weights.  If MS to pan weights are not provided,
        they are estimated from the data.  Uses the "How to Pan-sharpen Images Using the
        Gram-Schmidt Pan-sharpen Method – A Recipe" method described in
        https://doi.org/10.5194/isprsarchives-XL-1-W1-239-2013.
        """

        def get_weights(cov: np.ndarray, weights: np.ndarray = None) -> np.ndarray:
            """Return normalised MS to pan weights.  If ``weights`` is not provided, weights are
            estimated from the given pan / MS covariance.
            """
            if weights is None:
                # find the LS solution for the weights as described in section 2.2 & eq 6
                weights = np.linalg.lstsq(cov[1:, 1:], cov[0, 1:].reshape(-1, 1), rcond=None)[0]
                weights = weights.squeeze()

                # redo the LS without negatively weighted bands if there are any
                if np.any(weights < 0):
                    warnings.warn(
                        f'Weights contain negative value(s): {tuple(weights.round(4).tolist())}, '
                        f're-estimating positive weights.',
                        category=OrthorityWarning,
                    )
                    ms_indexes_ = np.where(weights > 0)[0] + 1
                    ms_cov = cov[ms_indexes_, :][:, ms_indexes_]
                    pan_ms_cov = cov[0, ms_indexes_].reshape(-1, 1)
                    weights_ = np.linalg.lstsq(ms_cov, pan_ms_cov, rcond=None)[0].squeeze()

                    # use the updated weights if they are positive, and set negative weights to 0
                    if np.all(weights_ >= 0):
                        weights = weights.clip(0, None)
                        weights[ms_indexes_ - 1] = weights_
            else:
                weights = np.array(weights)

            # set any remaining negative weights to 0, normalise & return
            weights = weights.flatten()
            if np.any(weights < 0):
                warnings.warn(
                    f'Weights contain negative value(s): {tuple(weights.round(4).tolist())}, '
                    f'setting to zero and normalising.',
                    category=OrthorityWarning,
                )
                weights = weights.clip(0, None)
            return weights / weights.sum()

        def get_gs_coeffs(cov: np.ndarray, weights: np.ndarray) -> list[np.ndarray]:
            """Return the Gram-Schmidt coefficients for the given MS covariance and MS to pan
            weights.
            """
            a = np.zeros((cov.shape[0], cov.shape[0]))
            coeffs = [np.zeros(k + 1) for k in range(cov.shape[0])]
            e = np.eye(cov.shape[0])
            for k in range(cov.shape[0]):
                # eq 4 from paper
                if k == 0:
                    a[k] = weights
                else:
                    a[k] = e[k - 1] - np.dot(coeffs[k - 1], a[:k])

                # eq 3 from paper
                for l in range(k + 1):
                    num = a[l].dot(cov[k])
                    den = (a[l].reshape(-1, 1).dot(a[l].reshape(1, -1)) * cov).sum()
                    # the 'if' below avoids dividing by zero with canonical weight vectors
                    coeffs[k][l] = num / den if np.any(a[l] != 0) else 0

            return coeffs

        def get_pan_norm(
            means: np.ndarray, cov: np.ndarray, weights: np.ndarray
        ) -> tuple[float, float]:
            """Return the gain and bias to convert the actual pan to the mean and standard
            deviation of the simulated pan.
            """
            # Note: this uses MS resolution, while the paper uses pan resolution actual pan
            # stats.  As the simulated pan stats are also at the MS resolution, this seems correct.

            pan_mean, pan_std = means[0], np.sqrt(cov[0, 0])

            # find simulated pan mean and std deviation from weights & MS covariances (no
            # corresponding equation in paper, derived from properties of covariance:
            # https://dlsun.github.io/probability/cov-properties.html)
            pan_sim_mean = weights.dot(means[1:])
            pan_sim_std = np.sqrt(
                (weights.reshape(-1, 1).dot(weights.reshape(1, -1)) * cov[1:, 1:]).sum()
            )

            # equations 2 & 3 from https://patents.google.com/patent/US6011875A/en
            gain = pan_sim_std / pan_std
            bias = pan_sim_mean - (gain * pan_mean)
            return gain, bias

        weights = get_weights(cov, weights=weights)
        coeffs = get_gs_coeffs(cov[1:, 1:], weights)
        gain, bias = get_pan_norm(means, cov, weights)

        if logger.getEffectiveLevel() <= logging.DEBUG:
            logger.debug(f"Pan / multispectral means: {means.round(4)}.")
            logger.debug(f"Pan / multispectral covariance: \n{cov.round(4)}.")
            logger.debug(f"Multispectral to pan weights: {weights.round(4).tolist()}.")
            coeffs_str = '\n'.join([str(c.round(4).tolist()) for c in coeffs])
            logger.debug(f"Gram-Schmidt coefficients: \n{coeffs_str}")
            logger.debug(f"Simulated pan gain: {gain:.4f}, bias: {bias:.4f}.")
        return dict(means=means, coeffs=coeffs, weights=weights, gain=gain, bias=bias)

    def _process_tile_array(
        self,
        pan_array: np.ndarray,
        ms_array: np.ndarray,
        means: np.ndarray,
        coeffs: list[np.ndarray],
        weights: np.ndarray,
        gain: float,
        bias: float,
    ) -> np.ndarray:
        """Return the pan-sharpened tile for the given pan & MS tiles, and Gram-Schmidt
        parameters.  Uses the "Process for Enhancing the Spatial Resolution of Multispectral
        Imagery Using Pan-sharpening" method in https://patents.google.com/patent/US6011875A/en.
        """
        # Note: the Gram-Schmidt transform is applied to upsampled MS data, rather than
        # transforming first then upsampling as in the patent.  This is equivalent and simpler in
        # code.

        def gs_foward(
            ms_array: np.ndarray,
            means: np.ndarray,
            coeffs: list[np.ndarray],
            weights: np.ndarray,
        ) -> np.ndarray:
            """Forward Gram-Schmidt transform of the given MS array with given parameters."""
            # equations 10-12 of the patent
            gs_array = np.zeros(
                (ms_array.shape[0] + 1, ms_array.shape[1]), dtype=self._working_dtype
            )
            gs_array[0] = weights.dot(ms_array)
            for bi in range(0, ms_array.shape[0]):
                phi = coeffs[bi]
                gs_array[bi + 1] = ms_array[bi] - means[bi] - phi.dot(gs_array[: bi + 1])

            return gs_array

        def gs_reverse(
            gs_array: np.ndarray,
            means: np.ndarray,
            coeffs: list[np.ndarray],
            ms_array: np.ndarray | None = None,
        ) -> np.ndarray:
            """Reverse Gram-Schmidt transform of the given Gram-Schmidt array with given
            parameters.  Optionally writes into ``ms_array`` if provided.
            """
            if ms_array is None:
                ms_array = np.zeros(
                    (gs_array.shape[0] - 1, gs_array.shape[1]), dtype=self._working_dtype
                )

            # equation 14 of the patent
            for bi in range(0, ms_array.shape[0]):
                phi = coeffs[bi]
                ms_array[bi] = gs_array[bi + 1] + means[bi] + phi.dot(gs_array[: bi + 1])

            return ms_array

        gs_array = gs_foward(ms_array, means[1:], coeffs, weights)
        gs_array[0] = (gain * pan_array) + bias  # substitute normalised pan
        ms_array = gs_reverse(gs_array, means[1:], coeffs, ms_array=ms_array)
        return ms_array

    def _process_tile(
        self,
        pan_im: rio.DatasetReader,
        ms_im: rio.DatasetReader,
        tile_win: Window,
        pan_index: int,
        ms_indexes: Sequence[int],
        out_im: rio.io.DatasetWriter,
        write_mask: bool,
        **params,
    ) -> None:
        """Thread-safe function to pan-sharpen a MS tile and write it to a dataset."""
        # read pan and MS tiles (without masked arrays which don't release GIL)
        with self._pan_lock:
            pan_array = pan_im.read(indexes=pan_index, window=tile_win)
        with self._ms_lock:
            ms_array = ms_im.read(indexes=ms_indexes, window=tile_win)

        # mask and reshape pan & MS
        pan_mask = np.logical_not(common.nan_equals(pan_array, pan_im.nodata))
        ms_mask = np.logical_not(common.nan_equals(ms_array, ms_im.nodata)).all(axis=0)
        mask = pan_mask & ms_mask
        pan_array_ = pan_array[mask].reshape(1, -1)
        ms_array_ = ms_array[:, mask].reshape(len(ms_indexes), -1)

        # pan-sharpen masked data and write into output mask area
        out_array_ = self._process_tile_array(pan_array_, ms_array_, **params)
        out_array_ = common.convert_array_dtype(out_array_, out_im.dtypes[0])
        out_array = np.full(ms_array.shape, fill_value=out_im.nodata or 0, dtype=out_im.dtypes[0])
        out_array[:, mask] = out_array_

        # write pan-sharpened tile
        with self._out_lock:
            out_im.write(out_array, window=tile_win)
            if write_mask:
                out_im.write_mask(mask, window=tile_win)

    def process(
        self,
        out_file: str | PathLike | OpenFile,
        pan_index: int = _default_alg_config['pan_index'],
        ms_indexes: Sequence[int] = _default_alg_config['ms_indexes'],
        weights: bool | Sequence[float] = _default_alg_config['weights'],
        interp: str | Interp = _default_alg_config['interp'],
        write_mask: bool | None = common._default_out_config['write_mask'],
        dtype: str = common._default_out_config['dtype'],
        compress: str | Compress | None = common._default_out_config['compress'],
        build_ovw: bool = common._default_out_config['build_ovw'],
        creation_options: dict | None = None,
        driver: str | Driver = common._default_out_config['driver'],
        overwrite: bool = common._default_out_config['overwrite'],
        progress: bool | Sequence[dict] = False,
    ):
        """
        Pan-sharpen.

        The pan-sharpened image is created on the panchromatic pixel grid.  Pan-sharpened image
        bounds are the intersection of the panchromatic and multispectral image bounds.

        Pan-sharpening consists of two steps, both of which operate tile-by-tile:

        1. Derive the Gram-Schmidt parameters from image statistics.
        2. Generate the pan-sharpened image.

        :param out_file:
            Pan-sharpened image file to create.  Can be a path or URI string, or an
            :class:`~fsspec.core.OpenFile` object in binary mode (``'wb'``).
        :param pan_index:
            Index of the panchromatic band to use (1-based).
        :param ms_indexes:
            Indexes of the multispectral bands to use (1-based).  If set to ``None`` (the default),
            all non-alpha multispectral bands are used.
        :param weights:
            Multi-spectral to panchromatic weights (≥0).  If set to ``None`` (the default),
            weights are estimated from the images.
        :param interp:
            Interpolation method for upsampling the multispectral image.
        :param write_mask:
            Mask valid pan-sharpened pixels with an internal mask (``True``), or with a nodata
            value based on ``dtype`` (``False``). An internal mask helps remove nodata noise
            caused by lossy compression. If set to ``None`` (the default), the mask will be
            written when JPEG compression is used.
        :param dtype:
            Pan-sharpened image data type (``uint8``, ``uint16``, ``int16``, ``float32`` or
            ``float64``).  If set to ``None`` (the default), the source image data type is used.
        :param compress:
            Pan-sharpened image compression type (``jpeg``, ``deflate`` or ``lzw``).  ``deflate``
            and ``lzw`` can be used with any ``dtype``, and ``jpeg`` with the uint8 ``dtype``.
            With supporting Rasterio builds, ``jpeg`` can also be used with uint16, in which case
            the ortho is 12 bit JPEG compressed.  If ``compress`` is set to ``None`` (the
            default), ``jpeg`` is used for the uint8 ``dtype``, and ``deflate`` otherwise.
        :param build_ovw:
            Whether to build overviews for the pan-sharpened image.
        :param creation_options:
            Pan-sharpened image creation options as dictionary of ``name: value`` pairs.  If
            supplied, ``compress`` is ignored.  See the `GDAL docs
            <https://gdal.org/en/latest/drivers/raster/gtiff.html#creation-options>`__ for details.
        :param driver:
            Pan-sharpened image driver (``gtiff`` or ``cog``).
        :param overwrite:
            Whether to overwrite the pan-sharpened image if it exists.
        :param progress:
            Whether to display a progress bar monitoring the portion of tiles processed in each
            step. Can be set to a sequence of two argument dictionaries that define a custom
            `tqdm <https://tqdm.github.io/docs/tqdm/>`_ bar for each step.
        """
        # set up progress bars
        if progress is True:
            progress = [
                common.get_tqdm_kwargs(desc=desc, unit='blocks')
                for desc in ['Statistics', 'Sharpening']
            ]
        elif progress is False:
            progress = [dict(disable=True, leave=False)] * 2

        exit_stack = ExitStack()
        interp = Interp(interp).to_rio()
        with exit_stack:
            exit_stack.enter_context(common.suppress_no_georef())

            # open pan & MS images
            exit_stack.enter_context(rio.Env(GDAL_NUM_THREADS='ALL_CPUS', GTIFF_FORCE_RGBA=False))
            pan_im = exit_stack.enter_context(common.OpenRaster(self._pan_file, 'r'))
            ms_im = exit_stack.enter_context(common.OpenRaster(self._ms_file, 'r'))

            # validate indexes and weights
            ms_indexes, weights = self._validate_pan_ms_params(
                pan_im, ms_im, pan_index, ms_indexes, weights
            )

            # open output image & resolve write_mask
            dtype = dtype or ms_im.dtypes[0]
            pan_profile = self._profiles['pan_to_pan']
            out_profile, write_mask = common.create_profile(
                driver=driver,
                shape=(len(ms_indexes), pan_profile['height'], pan_profile['width']),
                dtype=dtype,
                compress=compress,
                write_mask=write_mask,
                creation_options=creation_options,
            )
            out_profile.update(crs=pan_profile['crs'], transform=pan_profile['transform'])
            out_im = exit_stack.enter_context(
                common.OpenRaster(out_file, 'w', overwrite=overwrite, **out_profile)
            )

            # copy colorinterp from MS to output
            out_im.colorinterp = [ms_im.colorinterp[mi - 1] for mi in ms_indexes]

            # find pan-sharpening parameters from image stats
            means, cov = self._get_stats(pan_im, ms_im, pan_index, ms_indexes, progress[0])
            params = self._get_params(means, cov, weights)

            # open pan & upsampled MS WarpedVRTs that lie on the pan resolution grid, and have
            # common bounds
            pan_im = exit_stack.enter_context(WarpedVRT(pan_im, **self._profiles['pan_to_pan']))
            ms_im = exit_stack.enter_context(
                WarpedVRT(ms_im, **self._profiles['ms_to_pan'], resampling=interp),
            )

            # pan-sharpen tiles in a thread pool
            executor = exit_stack.enter_context(ThreadPoolExecutor(max_workers=os.cpu_count()))
            futures = [
                executor.submit(
                    self._process_tile,
                    pan_im,
                    ms_im,
                    tile_win,
                    pan_index,
                    ms_indexes,
                    out_im,
                    write_mask,
                    **params,
                )
                for tile_win in common.block_windows(out_im)
            ]

            pbar = exit_stack.enter_context(tqdm(**progress[1], total=len(futures)))
            for future in futures:
                try:
                    future.result()
                except Exception as ex:
                    # TODO: add cancel_futures=True to all thread pool shutdowns when supported py
                    #  versions are > 3.8
                    executor.shutdown(wait=False)
                    raise RuntimeError('Could not process tile.') from ex
                pbar.update()
            pbar.refresh()

            if build_ovw:
                common.build_overviews(out_im)



# Copyright The Orthority Contributors.
#
# This file is part of Orthority.
#
# Orthority is free software: you can redistribute it and/or modify it under the terms of the GNU
# Affero General Public License as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# Orthority is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Orthority.
# If not, see <https://www.gnu.org/licenses/>.
"""Camera model fitting and refinement."""
from __future__ import annotations

import logging
from copy import deepcopy
from typing import Sequence

import numpy as np
from rasterio.rpc import RPC

from orthority.camera import RpcCamera
from orthority.enums import RpcRefine

logger = logging.getLogger(__name__)

_default_rpc_refine_method = RpcRefine.shift


def refine_rpc(
    rpc: RPC | dict, gcps: Sequence[dict], method: RpcRefine = _default_rpc_refine_method
) -> dict:
    """
    Refine an RPC model with GCPs.

    Finds the least squares solution to the 'shift' or 'shift-and-drift' bias compensation
    refinements described in https://doi.org/10.1016/j.isprsjprs.2005.11.001 for a single image.
    Refinements are incorporated into the provided RPC model.  The approach is suited to narrow
    field of view satellite imagery.

    :param rpc:
        RPC parameters as a :class:`~rasterio.rpc.RPC` object or dictionary.
    :param gcps:
        List of GCP dictionaries, e.g. an item value in a dictionary returned by
        :func:`~orthority.param_io.read_im_gcps` or :func:`~orthority.param_io.read_oty_gcps`.
    :param method:
        Refinement method.

    :return:
        Refined RPC parameters as a dictionary.
    """
    # TODO:
    #  - robustness to outliers, maybe with optional inclusion of scipy
    #  - full affine model, or the method described in https://doi.org/10.14358/PERS.75.9.1083
    #  - fit evaluation with warnings
    method = RpcRefine(method)
    camera = RpcCamera(None, rpc)
    rpc = rpc.to_dict() if isinstance(rpc, RPC) else rpc
    min_gcps = 1 if method is RpcRefine.shift else 2
    if len(gcps) < min_gcps:
        raise ValueError(f"At least {min_gcps} are required for the '{method}' method.")

    def _norm_ji(rpc: dict, ji: np.ndarray) -> np.ndarray:
        """Normalise pixel coordinates with the given RPC scale / offset parameters."""
        norm_ji = ji.T - (rpc['samp_off'], rpc['line_off'])
        norm_ji /= (rpc['samp_scale'], rpc['line_scale'])
        return norm_ji.T

    # normalised GCP pixel coordinates
    ji_gcp = np.array([gcp['ji'] for gcp in gcps]).T
    ji_gcp = _norm_ji(rpc, ji_gcp)

    # normalised RPC pixel coordinates of GCP geographic coordinates
    xyz = np.array([(gcp['xyz']) for gcp in gcps]).T
    ji_rpc = camera.world_to_pixel(xyz)
    ji_rpc = _norm_ji(rpc, ji_rpc)

    # find the (partial) affine transform to "refine" RPC pixel coordinates
    refine_tform = np.eye(2, 3)
    if method == RpcRefine.shift:
        off = ji_gcp - ji_rpc
        refine_tform[:, -1] = off.mean(axis=1)
    else:
        for axis in range(2):
            ji_rpc_ = np.vstack((ji_rpc[axis], np.ones((ji_rpc.shape[1]))))
            (m, c), res, rank, s = np.linalg.lstsq(ji_rpc_.T, ji_gcp[axis], rcond=None)
            refine_tform[axis, axis] = m
            refine_tform[axis, 2] = c

    if logger.getEffectiveLevel() <= logging.DEBUG:
        # log the refinement transform and accuracy
        ji_rpc_ = np.vstack((ji_rpc, np.ones(ji_rpc.shape[1])))
        ji_refine = refine_tform.dot(ji_rpc_)
        err = ((ji_gcp - ji_refine).T * (rpc['samp_scale'], rpc['line_scale'])).T  # pixels
        err_dist = np.sum(err**2, axis=0)
        logger.debug(f"Refinement transform: \n{refine_tform}")
        logger.debug(f"Refinement RMSE (pixels): {np.sqrt(np.mean(err_dist)):.4f}")
        logger.debug(
            f"Refinement min - max error (pixels): "
            f"{np.sqrt(np.min(err_dist)):.5f} - {np.sqrt(np.max(err_dist)):.4f}"
        )

    # incorporate the refinement transform into the original RPC coefficients
    refined_rpc = deepcopy(rpc)
    for axis, num_key, den_key in zip(
        range(2), ['samp_num_coeff', 'line_num_coeff'], ['samp_den_coeff', 'line_den_coeff']
    ):
        refined_rpc[num_key] = np.array(refined_rpc[num_key]) * refine_tform[axis, axis]
        refined_rpc[num_key] += np.array(refined_rpc[den_key]) * refine_tform[axis, 2]
        refined_rpc[num_key] = refined_rpc[num_key].tolist()
    return refined_rpc
